@@ -1,7 +1,7 @@
 using System;
 using System.IO;
 using System.Text;
-using System.Threading;
+using ST = System.Threading;
 using System.Configuration;
 using System.Globalization;
 using System.Reflection;
@@ -14,11 +14,12 @@ using Mono.Debugger.Languages;
 
 namespace Mono.Debugger.Backend
 {
-	internal delegate void ChildOutputHandler (string output);
+	internal delegate void ChildOutputHandler (bool is_stderr, string output);
 
 	internal class Inferior : TargetAccess, ITargetNotification, IDisposable
 	{
 		protected IntPtr server_handle;
+		protected IntPtr io_data;
 		protected Bfd bfd;
 		protected ThreadManager thread_manager;
 
@@ -32,9 +33,6 @@ namespace Mono.Debugger.Backend
 		protected readonly AddressDomain address_domain;
 		protected readonly bool native;
 
-		protected ChildOutputHandler stdout_handler;
-		protected ChildOutputHandler stderr_handler;
-
 		int child_pid;
 		bool initialized;
 		bool has_target;
@@ -66,7 +64,9 @@ namespace Mono.Debugger.Backend
 		static extern TargetError mono_debugger_server_initialize_thread (IntPtr handle, int child_pid);
 
 		[DllImport("monodebuggerserver")]
-		static extern TargetError mono_debugger_server_spawn (IntPtr handle, string working_directory, string[] argv, string[] envp, out int child_pid, ChildOutputHandler stdout_handler, ChildOutputHandler stderr_handler, out IntPtr error);
+		static extern TargetError mono_debugger_server_io_thread_main (IntPtr io_data, ChildOutputHandler output_handler);
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_spawn (IntPtr handle, string working_directory, string[] argv, string[] envp, out int child_pid, out IntPtr io_data, out IntPtr error);
 
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_attach (IntPtr handle, int child_pid);
@@ -107,6 +107,9 @@ namespace Mono.Debugger.Backend
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_call_method_2 (IntPtr handle, long method_address, int data_size, IntPtr data, long callback_argument);
 
+		[DllImport("monodebuggerserver")]
+		static extern TargetError mono_debugger_server_call_method_3 (IntPtr handle, long method_address, long method_argument, long address_argument, int blob_size, IntPtr blob_data, long callback_argument);
+
 		[DllImport("monodebuggerserver")]
 		static extern TargetError mono_debugger_server_mark_rti_frame (IntPtr handle);
 
@@ -253,12 +256,13 @@ namespace Mono.Debugger.Backend
 			public int SIGINT;
 			public int SIGCHLD;
 
+			public int Kernel_SIGRTMIN;
 			public int MonoThreadAbortSignal;
 
 			public override string ToString ()
 			{
-				return String.Format ("SignalInfo ({0}:{1}:{2}:{3} - {4})",
-						      SIGKILL, SIGSTOP, SIGINT, SIGCHLD,
+				return String.Format ("SignalInfo ({0}:{1}:{2}:{3}:{4} - {5})",
+						      SIGKILL, SIGSTOP, SIGINT, SIGCHLD, Kernel_SIGRTMIN,
 						      MonoThreadAbortSignal);
 			}
 		}
@@ -377,6 +381,35 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
+		public void CallMethod (TargetAddress method, long method_argument,
+					TargetObject obj, long callback_arg)
+		{
+			check_disposed ();
+
+			byte[] blob = null;
+			long address = 0;
+
+			if (obj.Location.HasAddress)
+				address = obj.Location.GetAddress (this).Address;
+			else
+				blob = obj.Location.ReadBuffer (this, obj.Type.Size);
+
+			IntPtr blob_data = IntPtr.Zero;
+			try {
+				if (blob != null) {
+					blob_data = Marshal.AllocHGlobal (blob.Length);
+					Marshal.Copy (blob, 0, blob_data, blob.Length);
+				}
+
+				check_error (mono_debugger_server_call_method_3 (
+					server_handle, method.Address, method_argument,
+					address, blob != null ? blob.Length : 0, blob_data, callback_arg));
+			} finally {
+				if (blob_data != IntPtr.Zero)
+					Marshal.FreeHGlobal (blob_data);
+			}
+		}
+
 		public void RuntimeInvoke (TargetAddress invoke_method,
 					   TargetAddress method_argument,
 					   TargetObject object_argument,
@@ -550,7 +583,12 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
-		public int Run (bool redirect_fds)
+		void io_thread_main ()
+		{
+			mono_debugger_server_io_thread_main (io_data, process.OnTargetOutput);
+		}
+
+		public int Run ()
 		{
 			if (has_target)
 				throw new TargetException (TargetError.AlreadyHaveTarget);
@@ -559,15 +597,9 @@ namespace Mono.Debugger.Backend
 
 			IntPtr error;
 
-			stdout_handler = new ChildOutputHandler (inferior_stdout_handler);
-			stderr_handler = new ChildOutputHandler (inferior_stderr_handler);
-
 			TargetError result = mono_debugger_server_spawn (
 				server_handle, start.WorkingDirectory, start.CommandLineArguments,
-				start.Environment, out child_pid,
-				stdout_handler,
-				stderr_handler,
-				out error);
+				start.Environment, out child_pid, out io_data, out error);
 			if (result != TargetError.None) {
 				string message = Marshal.PtrToStringAuto (error);
 				g_free (error);
@@ -576,6 +608,10 @@ namespace Mono.Debugger.Backend
 					TargetError.CannotStartTarget, message);
 			}
 
+			ST.Thread io_thread = new ST.Thread (new ST.ThreadStart (io_thread_main));
+			io_thread.IsBackground = true;
+			io_thread.Start ();
+
 			return child_pid;
 		}
 
@@ -783,18 +819,6 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
-		void inferior_stdout_handler (string line)
-		{
-			if (TargetOutput != null)
-				TargetOutput (false, line);
-		}
-
-		void inferior_stderr_handler (string line)
-		{
-			if (TargetOutput != null)
-				TargetOutput (true, line);
-		}
-
 		public override int TargetIntegerSize {
 			get {
 				return target_info.TargetIntegerSize;
@@ -1049,7 +1073,6 @@ namespace Mono.Debugger.Backend
 		// IInferior
 		//
 
-		public event TargetOutputHandler TargetOutput;
 		public event StateChangedHandler StateChanged;
 
 		TargetState target_state = TargetState.NoTarget;
@@ -1473,6 +1496,10 @@ namespace Mono.Debugger.Backend
 			// child_event (ChildEventType.CHILD_MEMORY_CHANGED, 0);
 		}
 
+		public bool HasSignals {
+			get { return has_signals; }
+		}
+
 		public int SIGKILL {
 			get {
 				if (!has_signals || (signal_info.SIGKILL < 0))
@@ -1509,6 +1536,19 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
+		/*
+		 * CAUTION: This is the hard limit of the Linux kernel, not the first
+		 *          user-visible real-time signal !
+		 */
+		public int Kernel_SIGRTMIN {
+			get {
+				if (!has_signals || (signal_info.Kernel_SIGRTMIN < 0))
+					throw new InvalidOperationException ();
+
+				return signal_info.Kernel_SIGRTMIN;
+			}
+		}
+
 		public int MonoThreadAbortSignal {
 			get {
 				if (!has_signals || (signal_info.MonoThreadAbortSignal < 0))
