@@ -7,6 +7,7 @@ using System.Globalization;
 using System.Reflection;
 using System.Diagnostics;
 using System.Collections;
+using System.Collections.Generic;
 using System.Collections.Specialized;
 using System.Runtime.InteropServices;
 
@@ -46,15 +47,16 @@ namespace Mono.Debugger.Backend
 		DomainCreate,
 		DomainUnload,
 		ClassInitialized,
+		InterruptionRequest,
 
 		Trampoline	= 256
 	}
 
+	internal delegate object ManagedCallbackFunction (SingleSteppingEngine engine);
+
 	internal class MonoThreadManager
 	{
-		ThreadManager thread_manager;
 		MonoDebuggerInfo debugger_info;
-		Inferior inferior;
 
 		public static MonoThreadManager Initialize (ThreadManager thread_manager,
 							    Inferior inferior, bool attach)
@@ -73,11 +75,10 @@ namespace Mono.Debugger.Backend
 		protected MonoThreadManager (ThreadManager thread_manager, Inferior inferior,
 					     TargetAddress info, bool attach)
 		{
-			this.inferior = inferior;
-			this.thread_manager = thread_manager;
-
 			debugger_info = MonoDebuggerInfo.Create (inferior, info);
 
+			inferior.WriteInteger (debugger_info.UsingMonoDebugger, 1);
+
 			notification_bpt = new InitializeBreakpoint (this, debugger_info.Initialize);
 			notification_bpt.Insert (inferior);
 		}
@@ -85,6 +86,11 @@ namespace Mono.Debugger.Backend
 		AddressBreakpoint notification_bpt;
 		IntPtr mono_runtime_info;
 
+		internal bool HasCodeBuffer {
+			get;
+			private set;
+		}
+
 		[DllImport("monodebuggerserver")]
 		static extern IntPtr mono_debugger_server_initialize_mono_runtime (
 			int address_size, long notification_address,
@@ -95,16 +101,20 @@ namespace Mono.Debugger.Backend
 		[DllImport("monodebuggerserver")]
 		static extern void mono_debugger_server_finalize_mono_runtime (IntPtr handle);
 
+		[DllImport("monodebuggerserver")]
+		static extern void mono_debugger_server_initialize_code_buffer (
+			IntPtr runtime, long executable_code_buffer,
+			int executable_code_buffer_size);
+
 		protected void initialize_notifications (Inferior inferior)
 		{
-			TargetAddress notification_address = inferior.ReadAddress (
-				debugger_info.NotificationAddress);
 			TargetAddress executable_code_buffer = inferior.ReadAddress (
 				debugger_info.ExecutableCodeBuffer);
+			HasCodeBuffer = !executable_code_buffer.IsNull;
 
 			mono_runtime_info = mono_debugger_server_initialize_mono_runtime (
 				inferior.TargetAddressSize,
-				notification_address.Address,
+				debugger_info.NotificationAddress.Address,
 				executable_code_buffer.Address,
 				debugger_info.ExecutableCodeBufferSize,
 				debugger_info.BreakpointInfo.Address,
@@ -112,7 +122,7 @@ namespace Mono.Debugger.Backend
 				debugger_info.BreakpointArraySize);
 			inferior.SetRuntimeInfo (mono_runtime_info);
 
-			inferior.WriteInteger (debugger_info.DebuggerVersion, 3);
+			inferior.WriteInteger (debugger_info.DebuggerVersion, 4);
 
 			if (notification_bpt != null) {
 				notification_bpt.Remove (inferior);
@@ -120,6 +130,14 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
+		internal void InitCodeBuffer (Inferior inferior, TargetAddress code_buffer)
+		{
+			HasCodeBuffer = true;
+			mono_debugger_server_initialize_code_buffer (
+				mono_runtime_info, code_buffer.Address,
+				debugger_info.ExecutableCodeBufferSize);
+		}
+
 		protected class InitializeBreakpoint : AddressBreakpoint
 		{
 			protected readonly MonoThreadManager manager;
@@ -144,9 +162,43 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
+		internal void InitializeAfterAttach (Inferior inferior)
+		{
+			initialize_notifications (inferior);
+
+			inferior.WriteAddress (debugger_info.ThreadVTablePtr,
+					       debugger_info.ThreadVTable);
+			inferior.WriteAddress (debugger_info.EventHandlerPtr,
+					       debugger_info.EventHandler);
+			inferior.WriteInteger (debugger_info.UsingMonoDebugger, 1);
+
+			csharp_language = inferior.Process.CreateMonoLanguage (debugger_info);
+			csharp_language.InitializeAttach (inferior);
+			inferior.InitializeModules ();
+		}
+
+		internal void InitializeAfterExec (Inferior inferior)
+		{
+			inferior.WriteInteger (debugger_info.UsingMonoDebugger, 1);
+		}
+
+		internal void Detach (Inferior inferior)
+		{
+			inferior.WriteAddress (debugger_info.ThreadVTablePtr, TargetAddress.Null);
+			inferior.WriteAddress (debugger_info.EventHandler, TargetAddress.Null);
+			inferior.WriteInteger (debugger_info.UsingMonoDebugger, 0);
+		}
+
+		internal void AddManagedCallback (Inferior inferior, ManagedCallbackFunction func)
+		{
+			inferior.WriteInteger (MonoDebuggerInfo.InterruptionRequest, 1);
+			managed_callbacks.Enqueue (func);
+		}
+
 		TargetAddress main_function;
 		TargetAddress main_thread;
 		MonoLanguageBackend csharp_language;
+		Queue<ManagedCallbackFunction> managed_callbacks = new Queue<ManagedCallbackFunction> ();
 
 		internal bool CanExecuteCode {
 			get { return mono_runtime_info != IntPtr.Zero; }
@@ -246,6 +298,14 @@ namespace Mono.Debugger.Backend
 
 					break;
 
+				case NotificationType.ReachedMain: {
+					Inferior.StackFrame iframe = inferior.GetCurrentFrame (false);
+					engine.SetMainReturnAddress (iframe.StackPointer);
+					engine.ProcessServant.OnProcessReachedMainEvent ();
+					resume_target = !engine.OnModuleLoaded ();
+					return true;
+				}
+
 				case NotificationType.WrapperMain:
 				case NotificationType.MainExited:
 					break;
@@ -284,6 +344,15 @@ namespace Mono.Debugger.Backend
 				case NotificationType.ClassInitialized:
 					break;
 
+#if MARTIN_PRIVATE
+				case NotificationType.InterruptionRequest:
+					inferior.WriteInteger (MonoDebuggerInfo.InterruptionRequest, 0);
+					engine.DoManagedCallback (managed_callbacks.ToArray ());
+					managed_callbacks = new Queue<ManagedCallbackFunction> ();
+					resume_target = false;
+					return true;
+#endif
+
 				default: {
 					TargetAddress data = new TargetAddress (
 						inferior.AddressDomain, cevent.Data1);
@@ -316,8 +385,8 @@ namespace Mono.Debugger.Backend
 	internal class MonoDebuggerInfo
 	{
 		// These constants must match up with those in mono/mono/metadata/mono-debug.h
-		public const int  MinDynamicVersion = 66;
-		public const int  MaxDynamicVersion = 66;
+		public const int  MinDynamicVersion = 80;
+		public const int  MaxDynamicVersion = 80;
 		public const long DynamicMagic      = 0x7aff65af4253d427;
 
 		public readonly int MonoTrampolineNum;
@@ -340,20 +409,27 @@ namespace Mono.Debugger.Backend
 		public readonly TargetAddress RemoveBreakpoint;
 		public readonly TargetAddress RegisterClassInitCallback;
 		public readonly TargetAddress RemoveClassInitCallback;
-		public readonly TargetAddress Attach;
-		public readonly TargetAddress Detach;
 		public readonly TargetAddress Initialize;
-		public readonly TargetAddress GetLMFAddress;
 		public readonly TargetAddress ThreadTable;
 		public readonly TargetAddress ExecutableCodeBuffer;
 		public readonly TargetAddress BreakpointInfo;
 		public readonly TargetAddress BreakpointInfoIndex;
 		public readonly int ExecutableCodeBufferSize;
 		public readonly int BreakpointArraySize;
+		public readonly TargetAddress GetMethodSignature;
+		public readonly TargetAddress InitCodeBuffer;
+
+		public readonly TargetAddress ThreadVTablePtr;
+		public readonly TargetAddress ThreadVTable;
+		public readonly TargetAddress EventHandlerPtr;
+		public readonly TargetAddress EventHandler;
+
+		public readonly TargetAddress UsingMonoDebugger;
+		public readonly TargetAddress InterruptionRequest;
 
 		public static MonoDebuggerInfo Create (TargetMemoryAccess memory, TargetAddress info)
 		{
-			TargetBinaryReader header = memory.ReadMemory (info, 16).GetReader ();
+			TargetBinaryReader header = memory.ReadMemory (info, 24).GetReader ();
 			long magic = header.ReadInt64 ();
 			if (magic != DynamicMagic)
 				throw new SymbolTableException (
@@ -371,6 +447,9 @@ namespace Mono.Debugger.Backend
 					"but expected at most {1}.", version,
 					MonoDebuggerInfo.MaxDynamicVersion);
 
+			int minor_version = header.ReadInt32 ();
+			header.ReadInt32 ();
+
 			int size = header.ReadInt32 ();
 
 			TargetReader reader = new TargetReader (memory.ReadMemory (info, size));
@@ -380,7 +459,7 @@ namespace Mono.Debugger.Backend
 		protected MonoDebuggerInfo (TargetMemoryAccess memory, TargetReader reader)
 		{
 			/* skip past magic, version, and total_size */
-			reader.Offset = 16;
+			reader.Offset = 24;
 
 			SymbolTableSize           = reader.ReadInteger ();
 			MonoTrampolineNum         = reader.ReadInteger ();
@@ -396,10 +475,7 @@ namespace Mono.Debugger.Backend
 			RuntimeInvoke             = reader.ReadAddress ();
 			ClassGetStaticFieldData   = reader.ReadAddress ();
 			RunFinally                = reader.ReadAddress ();
-			Attach                    = reader.ReadAddress ();
-			Detach                    = reader.ReadAddress ();
 			Initialize                = reader.ReadAddress ();
-			GetLMFAddress             = reader.ReadAddress ();
 
 			CreateString              = reader.ReadAddress ();
 			LookupClass               = reader.ReadAddress ();
@@ -420,6 +496,17 @@ namespace Mono.Debugger.Backend
 			ExecutableCodeBufferSize  = reader.ReadInteger ();
 			BreakpointArraySize       = reader.ReadInteger ();
 
+			GetMethodSignature        = reader.ReadAddress ();
+			InitCodeBuffer            = reader.ReadAddress ();
+
+			ThreadVTablePtr           = reader.ReadAddress ();
+			ThreadVTable              = reader.ReadAddress ();
+			EventHandlerPtr           = reader.ReadAddress ();
+			EventHandler              = reader.ReadAddress ();
+
+			UsingMonoDebugger         = reader.ReadAddress ();
+			InterruptionRequest       = reader.ReadAddress ();
+
 			Report.Debug (DebugFlags.JitSymtab, this);
 		}
 	}
