@@ -1,6 +1,7 @@
 using System;
 using System.IO;
 using System.Collections;
+using System.Collections.Generic;
 using ST = System.Threading;
 using System.Runtime.Serialization;
 using System.Runtime.Serialization.Formatters.Binary;
@@ -31,6 +32,7 @@ namespace Mono.Debugger.Backend
 		ThreadDB thread_db;
 
 		bool is_attached;
+		bool is_execed;
 		bool is_forked;
 		bool initialized;
 		ST.ManualResetEvent initialized_event;
@@ -97,6 +99,10 @@ namespace Mono.Debugger.Backend
 			get { return is_attached; }
 		}
 
+		public bool IsExeced {
+			get { return is_execed; }
+		}
+
 		public Process Client {
 			get { return client; }
 		}
@@ -187,7 +193,7 @@ namespace Mono.Debugger.Backend
 			if ((mono_manager != null) && !do_attach)
 				mono_manager.ThreadCreated (new_thread);
 
-			if (!do_attach)
+			if (!do_attach && !is_execed)
 				get_thread_info (inferior, new_thread);
 			OnThreadCreatedEvent (new_thread);
 		}
@@ -223,6 +229,8 @@ namespace Mono.Debugger.Backend
 
 		internal void ChildExecd (Inferior inferior)
 		{
+			is_execed = true;
+
 			if (!is_forked) {
 				if (bfd_container != null)
 					bfd_container.Dispose ();
@@ -292,6 +300,16 @@ namespace Mono.Debugger.Backend
 			manager.Debugger.OnThreadCreatedEvent (thread.Client);
 		}
 
+		internal void OnManagedThreadExitedEvent (ThreadServant thread)
+		{
+			thread_hash.Remove (thread.PID);
+		}
+
+		internal void OnProcessReachedMainEvent ()
+		{
+			manager.Debugger.Client.OnProcessReachedMainEvent (client);
+		}
+
 		internal void OnThreadExitedEvent (ThreadServant thread)
 		{
 			thread_hash.Remove (thread.PID);
@@ -303,6 +321,11 @@ namespace Mono.Debugger.Backend
 				OnProcessExitedEvent ();
 		}
 
+		internal void OnTargetOutput (bool is_stderr, string output)
+		{
+			client.OnTargetOutput (is_stderr, output);
+		}
+
 		internal void WaitForApplication ()
 		{
 			initialized_event.WaitOne ();
@@ -321,6 +344,10 @@ namespace Mono.Debugger.Backend
 			if (!is_forked || is_exec) {
 				mono_manager = MonoThreadManager.Initialize (
 					manager, inferior, (start.PID != 0) && !is_exec);
+
+				if (is_exec && (mono_manager != null))
+					mono_manager.InitializeAfterExec (inferior);
+
 				if (!is_forked && !is_exec && !is_attached &&
 				    !start.IsNative && (mono_manager == null))
 					throw new TargetException (TargetError.CannotStartTarget,
@@ -356,6 +383,9 @@ namespace Mono.Debugger.Backend
 					read_thread_table (inferior);
 			}
 
+			if (!IsManaged)
+				manager.Debugger.Client.OnProcessReachedMainEvent (client);
+
 			initialized_event.Set ();
 		}
 
@@ -395,7 +425,7 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
-		internal void InitializeThreads (TargetMemoryAccess target)
+		internal void InitializeThreads (Inferior target)
 		{
 			thread_db = ThreadDB.Create (this, target);
 			if (thread_db == null) {
@@ -562,6 +592,11 @@ namespace Mono.Debugger.Backend
 			}
 		}
 
+		public TargetAddress LookupSymbol (string name)
+		{
+			return bfd_container.LookupSymbol (name);
+		}
+
 		public Thread[] GetThreads ()
 		{
 			lock (thread_hash.SyncRoot) {
@@ -628,6 +663,15 @@ namespace Mono.Debugger.Backend
 				      "Released global thread lock #1: {0}", caller);
 		}
 
+		//
+		// Stopping / resuming all threads for the GUI
+		//
+
+		internal void OnTargetEvent (SingleSteppingEngine sse, TargetEventArgs args)
+		{
+			client.OnTargetEvent (sse, args);
+		}
+
 		//
 		// IDisposable
 		//
