@@ -1,35 +1,13 @@
 using System;
 using System.Collections;
-using Mono.Debugger.Backends;
+using Mono.Debugger.Backend;
 
-namespace Mono.Debugger.Backends
+namespace Mono.Debugger.Architectures
 {
-	// Keep in sync with DebuggerI386Registers in backends/server/i386-arch.h.
-	internal enum I386Register
-	{
-		EBX		= 0,
-		ECX		= 1,
-		EDX		= 2,
-		ESI		= 3,
-		EDI		= 4,
-		EBP		= 5,
-		EAX		= 6,
-		DS		= 7,
-		ES		= 8,
-		FS		= 9,
-		GS		= 10,
-		EIP		= 11,
-		CS		= 12,
-		EFLAGS		= 13,
-		ESP		= 14,
-		SS		= 15,
-		COUNT		= 16
-	}
-
 	// <summary>
 	//   Architecture-dependent stuff for the i386.
 	// </summary>
-	internal class Architecture_I386 : Architecture
+	internal class Architecture_I386 : X86_Architecture
 	{
 		internal Architecture_I386 (ProcessServant process, TargetInfo info)
 			: base (process, info)
@@ -56,250 +34,96 @@ namespace Mono.Debugger.Backends
 			}
 		}
 
-		internal override TargetAddress GetCallTarget (TargetMemoryAccess target,
-							       TargetAddress address,
-							       out int insn_size)
-		{
-			if (address.Address == 0xffffe002) {
-				insn_size = 0;
-				return TargetAddress.Null;
-			}
-
-			TargetBinaryReader reader = target.ReadMemory (address, 6).GetReader ();
-
-			byte opcode = reader.ReadByte ();
-
-			if (opcode == 0xe8) {
-				int call_target = reader.ReadInt32 ();
-				insn_size = 5;
-				return address + reader.Position + call_target;
-			} else if (opcode != 0xff) {
-				insn_size = 0;
-				return TargetAddress.Null;
-			}
-
-			byte address_byte = reader.ReadByte ();
-			byte register;
-			int disp;
-			bool dereference_addr;
-
-			if (((address_byte & 0x38) == 0x10) && ((address_byte >> 6) == 1)) {
-				register = (byte) (address_byte & 0x07);
-				disp = reader.ReadByte ();
-				insn_size = 3;
-				dereference_addr = true;
-			} else if (((address_byte & 0x38) == 0x10) && ((address_byte >> 6) == 2)) {
-				register = (byte) (address_byte & 0x07);
-				disp = reader.ReadInt32 ();
-				insn_size = 6;
-				dereference_addr = true;
-			} else if (((address_byte & 0x38) == 0x10) && ((address_byte >> 6) == 3)) {
-				register = (byte) (address_byte & 0x07);
-				disp = 0;
-				insn_size = 2;
-				dereference_addr = false;
-			} else if (((address_byte & 0x38) == 0x10) && ((address_byte >> 6) == 0)) {
-				register = (byte) (address_byte & 0x07);
-				disp = 0;
-				insn_size = 2;
-				dereference_addr = true;
-			} else {
-				insn_size = 0;
-				return TargetAddress.Null;
-			}
-
-			I386Register reg;
-			switch (register) {
-			case 0:
-				reg = I386Register.EAX;
-				break;
-
-			case 1:
-				reg = I386Register.ECX;
-				break;
-
-			case 2:
-				reg = I386Register.EDX;
-				break;
-
-			case 3:
-				reg = I386Register.EBX;
-				break;
-
-			case 6:
-				reg = I386Register.ESI;
-				break;
-
-			case 7:
-				reg = I386Register.EDI;
-				break;
-
-			default:
-				insn_size = 0;
-				return TargetAddress.Null;
-			}
-
-			Registers regs = target.GetRegisters ();
-			Register addr = regs [(int) reg];
-
-			TargetAddress vtable_addr = new TargetAddress (AddressDomain, addr.Value);
-			vtable_addr += disp;
-
-			if (dereference_addr)
-				return target.ReadAddress (vtable_addr);
-			else
-				return vtable_addr;
-		}
-
-		internal override TargetAddress GetJumpTarget (TargetMemoryAccess target,
-							       TargetAddress address,
-							       out int insn_size)
-		{
-			TargetBinaryReader reader = target.ReadMemory (address, 10).GetReader ();
-
-			byte opcode = reader.ReadByte ();
-			byte opcode2 = reader.ReadByte ();
-
-			if ((opcode == 0xff) && (opcode2 == 0x25)) {
-				insn_size = 2 + TargetAddressSize;
-				return new TargetAddress (AddressDomain, reader.ReadAddress ());
-			} else if ((opcode == 0xff) && (opcode2 == 0xa3)) {
-				int offset = reader.ReadInt32 ();
-				Registers regs = target.GetRegisters ();
-				long ebx = regs [(int) I386Register.EBX].Value;
-
-				insn_size = 6;
-				return new TargetAddress (AddressDomain, ebx + offset);
-			}
-
-			insn_size = 0;
-			return TargetAddress.Null;
-		}
-
-		internal override TargetAddress GetTrampoline (TargetMemoryAccess target,
-							       TargetAddress location,
-							       TargetAddress trampoline_address)
-		{
-			if (trampoline_address.IsNull)
-				return TargetAddress.Null;
-
-			TargetBinaryReader reader = target.ReadMemory (location, 10).GetReader ();
-
-			byte opcode = reader.ReadByte ();
-			if (opcode != 0x68)
-				return TargetAddress.Null;
-
-			int method_info = reader.ReadInt32 ();
-
-			opcode = reader.ReadByte ();
-			if (opcode != 0xe9)
-				return TargetAddress.Null;
-
-			int call_disp = reader.ReadInt32 ();
-
-			if (location + call_disp + 10 != trampoline_address)
-				return TargetAddress.Null;
-
-			return new TargetAddress (AddressDomain, method_info);
-		}
-
-		public override string[] RegisterNames {
+		public override int[] AllRegisterIndices {
 			get {
-				return registers;
+				return new int[] {
+					(int) X86_Register.RAX, (int) X86_Register.RCX,
+					(int) X86_Register.RDX, (int) X86_Register.RBX,
+					(int) X86_Register.RSP, (int) X86_Register.RBP,
+					(int) X86_Register.RSI, (int) X86_Register.RDI,
+
+					(int) X86_Register.RIP,
+					(int) X86_Register.EFLAGS,
+					(int) X86_Register.ORIG_RAX,
+
+					(int) X86_Register.CS,  (int) X86_Register.SS,
+					(int) X86_Register.DS,  (int) X86_Register.ES,
+					(int) X86_Register.FS,  (int) X86_Register.GS,
+
+					(int) X86_Register.FS_BASE,
+					(int) X86_Register.GS_BASE
+				};
 			}
 		}
 
 		public override int[] RegisterIndices {
 			get {
-				return important_regs;
-			}
-		}
+				return new int[] {
+					(int) X86_Register.RAX, (int) X86_Register.RCX,
+					(int) X86_Register.RDX, (int) X86_Register.RBX,
+					(int) X86_Register.RSP, (int) X86_Register.RBP,
+					(int) X86_Register.RSI, (int) X86_Register.RDI,
 
-		public override int[] AllRegisterIndices {
-			get {
-				return all_regs;
-			}
-		}
-
-		public override int[] RegisterSizes {
-			get {
-				return reg_sizes;
+					(int) X86_Register.RIP, (int) X86_Register.EFLAGS
+				};
 			}
 		}
 
+		// FIXME: Map mono/arch/x86/x86-codegen.h registers to
+		//        debugger/arch/IArchitecture_I386.cs registers.
 		internal override int[] RegisterMap {
 			get {
-				return register_map;
+				return new int[] {
+					(int) X86_Register.RAX, (int) X86_Register.RCX,
+					(int) X86_Register.RDX, (int) X86_Register.RBX,
+					(int) X86_Register.RSP, (int) X86_Register.RBP,
+					(int) X86_Register.RSI, (int) X86_Register.RDI
+				};
 			}
 		}
 
 		internal override int[] DwarfFrameRegisterMap {
 			get {
-				return dwarf_frame_register_map;
+				return new int[] {
+					(int) X86_Register.RIP, (int) X86_Register.RSP,
+					(int) X86_Register.RBP,
+
+					(int) X86_Register.RAX, (int) X86_Register.RBX,
+					(int) X86_Register.RCX, (int) X86_Register.RDX,
+					(int) X86_Register.RSP, (int) X86_Register.RBP,
+					(int) X86_Register.RSI, (int) X86_Register.RDI,
+				};
 			}
 		}
 
-		internal override int CountRegisters {
+
+		public override string[] RegisterNames {
 			get {
-				return (int) I386Register.COUNT;
+				return new string[] {
+					"eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi",
+					null, null, null, null, null, null, null, null,
+					"eip", "eflags", null, "cs", "ss", "ds", "es",
+					"fs", "gs", null, null
+				};
 			}
 		}
 
-		int[] all_regs = { (int) I386Register.EBX,
-				   (int) I386Register.ECX,
-				   (int) I386Register.EDX,
-				   (int) I386Register.ESI,
-				   (int) I386Register.EDI,
-				   (int) I386Register.EBP,
-				   (int) I386Register.EAX,
-				   (int) I386Register.DS,
-				   (int) I386Register.ES,
-				   (int) I386Register.FS,
-				   (int) I386Register.GS,
-				   (int) I386Register.EIP,
-				   (int) I386Register.CS,
-				   (int) I386Register.EFLAGS,
-				   (int) I386Register.ESP,
-				   (int) I386Register.SS };
-
-		int[] important_regs = { (int) I386Register.EAX,
-					 (int) I386Register.EBX,
-					 (int) I386Register.ECX,
-					 (int) I386Register.EDX,
-					 (int) I386Register.ESI,
-					 (int) I386Register.EDI,
-					 (int) I386Register.EBP,
-					 (int) I386Register.ESP,
-					 (int) I386Register.EIP,
-					 (int) I386Register.EFLAGS };
-
-		// FIXME: Map mono/arch/x86/x86-codegen.h registers to
-		//        debugger/arch/IArchitecture_I386.cs registers.
-		int[] register_map = { (int) I386Register.EAX, (int) I386Register.ECX,
-				       (int) I386Register.EDX, (int) I386Register.EBX,
-				       (int) I386Register.ESP, (int) I386Register.EBP,
-				       (int) I386Register.ESI, (int) I386Register.EDI };
-
-		int[] dwarf_frame_register_map = new int [] {
-			(int) I386Register.EIP, (int) I386Register.ESP, (int) I386Register.EBP,
-
-			(int) I386Register.EAX, (int) I386Register.EBX, (int) I386Register.ECX,
-			(int) I386Register.EDX, (int) I386Register.ESP, (int) I386Register.EBP,
-			(int) I386Register.ESI, (int) I386Register.EDI
-		};
-				
-		string[] registers = { "ebx", "ecx", "edx", "esi", "edi", "ebp", "eax", "ds",
-				       "es", "fs", "gs", "eip", "cs", "eflags", "esp", "ss" };
-
-		int[] reg_sizes = { 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4 };
+		public override int[] RegisterSizes {
+			get {
+				return new int[] {
+					4, 4, 4, 4, 4, 4, 4, 4, -1, -1, -1, -1, -1, -1, -1, -1,
+					4, 4, -1, 4, 4, 4, 4, 4, 4, -1, -1
+				};
+			}
+		}
 
 		public override string PrintRegister (Register register)
 		{
 			if (!register.Valid)
 				return "XXXXXXXX";
 
-			switch ((I386Register) register.Index) {
-			case I386Register.EFLAGS: {
+			switch ((X86_Register) register.Index) {
+			case X86_Register.EFLAGS: {
 				ArrayList flags = new ArrayList ();
 				long value = register.Value;
 				if ((value & (1 << 0)) != 0)
@@ -362,16 +186,16 @@ namespace Mono.Debugger.Backends
 			return String.Format (
 				"EAX={0}  EBX={1}  ECX={2}  EDX={3}  ESI={4}  EDI={5}\n" +
 				"EBP={6}  ESP={7}  EIP={8}  EFLAGS={9}\n",
-				format (registers [(int) I386Register.EAX]),
-				format (registers [(int) I386Register.EBX]),
-				format (registers [(int) I386Register.ECX]),
-				format (registers [(int) I386Register.EDX]),
-				format (registers [(int) I386Register.ESI]),
-				format (registers [(int) I386Register.EDI]),
-				format (registers [(int) I386Register.EBP]),
-				format (registers [(int) I386Register.ESP]),
-				format (registers [(int) I386Register.EIP]),
-				PrintRegister (registers [(int) I386Register.EFLAGS]));
+				format (registers [(int) X86_Register.RAX]),
+				format (registers [(int) X86_Register.RBX]),
+				format (registers [(int) X86_Register.RCX]),
+				format (registers [(int) X86_Register.RDX]),
+				format (registers [(int) X86_Register.RSI]),
+				format (registers [(int) X86_Register.RDI]),
+				format (registers [(int) X86_Register.RBP]),
+				format (registers [(int) X86_Register.RSP]),
+				format (registers [(int) X86_Register.RIP]),
+				PrintRegister (registers [(int) X86_Register.EFLAGS]));
 		}
 
 		internal override int MaxPrologueSize {
@@ -385,20 +209,20 @@ namespace Mono.Debugger.Backends
 			Registers regs = new Registers (old_regs);
 
 			TargetAddress ebp = new TargetAddress (
-				AddressDomain, old_regs [(int) I386Register.EBP].GetValue ());
+				memory.AddressDomain, old_regs [(int) X86_Register.RBP].GetValue ());
 
 			int addr_size = TargetAddressSize;
 			TargetAddress new_ebp = memory.ReadAddress (ebp);
-			regs [(int) I386Register.EBP].SetValue (ebp, new_ebp);
+			regs [(int) X86_Register.RBP].SetValue (ebp, new_ebp);
 
 			TargetAddress new_eip = memory.ReadAddress (ebp + addr_size);
-			regs [(int) I386Register.EIP].SetValue (ebp + addr_size, new_eip);
+			regs [(int) X86_Register.RIP].SetValue (ebp + addr_size, new_eip);
 
 			TargetAddress new_esp = ebp + 2 * addr_size;
-			regs [(int) I386Register.ESP].SetValue (ebp, new_esp);
+			regs [(int) X86_Register.RSP].SetValue (ebp, new_esp);
 
-			regs [(int) I386Register.ESI].Valid = true;
-			regs [(int) I386Register.EDI].Valid = true;
+			regs [(int) X86_Register.RSI].Valid = true;
+			regs [(int) X86_Register.RDI].Valid = true;
 
 			ebp -= addr_size;
 
@@ -413,34 +237,35 @@ namespace Mono.Debugger.Backends
 				switch (opcode) {
 				case 0x50: /* eax */
 					value = (long) (uint) memory.ReadInteger (ebp);
-					regs [(int) I386Register.EAX].SetValue (ebp, value);
+					regs [(int) X86_Register.RAX].SetValue (ebp, value);
 					break;
 				case 0x51: /* ecx */
 					value = (long) (uint) memory.ReadInteger (ebp);
-					regs [(int) I386Register.ECX].SetValue (ebp, value);
+					regs [(int) X86_Register.RCX].SetValue (ebp, value);
 					break;
 				case 0x52: /* edx */
 					value = (long) (uint) memory.ReadInteger (ebp);
-					regs [(int) I386Register.EDX].SetValue (ebp, value);
+					regs [(int) X86_Register.RDX].SetValue (ebp, value);
 					break;
 				case 0x53: /* ebx */
 					value = (long) (uint) memory.ReadInteger (ebp);
-					regs [(int) I386Register.EBX].SetValue (ebp, value);
+					regs [(int) X86_Register.RBX].SetValue (ebp, value);
 					break;
 				case 0x56: /* esi */
 					value = (long) (uint) memory.ReadInteger (ebp);
-					regs [(int) I386Register.ESI].SetValue (ebp, value);
+					regs [(int) X86_Register.RSI].SetValue (ebp, value);
 					break;
 				case 0x57: /* edi */
 					value = (long) (uint) memory.ReadInteger (ebp);
-					regs [(int) I386Register.EDI].SetValue (ebp, value);
+					regs [(int) X86_Register.RDI].SetValue (ebp, value);
 					break;
 				}
 
 				ebp -= addr_size;
 			}
 
-			return CreateFrame (frame.Thread, new_eip, new_esp, new_ebp, regs);
+			return CreateFrame (frame.Thread, memory, new_eip, new_esp,
+					    new_ebp, regs, true);
 		}
 
 		StackFrame read_prologue (StackFrame frame, TargetMemoryAccess memory,
@@ -466,14 +291,15 @@ namespace Mono.Debugger.Backends
 				Registers regs = new Registers (old_regs);
 
 				TargetAddress new_eip = memory.ReadAddress (frame.StackPointer);
-				regs [(int) I386Register.EIP].SetValue (frame.StackPointer, new_eip);
+				regs [(int) X86_Register.RIP].SetValue (frame.StackPointer, new_eip);
 
 				TargetAddress new_esp = frame.StackPointer + TargetAddressSize;
 				TargetAddress new_ebp = frame.FrameAddress;
 
-				regs [(int) I386Register.ESP].SetValue (new_esp);
+				regs [(int) X86_Register.RSP].SetValue (new_esp);
 
-				return CreateFrame (frame.Thread, new_eip, new_esp, new_ebp, regs);
+				return CreateFrame (frame.Thread, memory, new_eip, new_esp,
+						    new_ebp, regs, true);
 			}
 
 			// push %ebp
@@ -486,16 +312,17 @@ namespace Mono.Debugger.Backends
 
 				int addr_size = TargetAddressSize;
 				TargetAddress new_ebp = memory.ReadAddress (frame.StackPointer);
-				regs [(int) I386Register.EBP].SetValue (frame.StackPointer, new_ebp);
+				regs [(int) X86_Register.RBP].SetValue (frame.StackPointer, new_ebp);
 
 				TargetAddress new_esp = frame.StackPointer + addr_size;
 				TargetAddress new_eip = memory.ReadAddress (new_esp);
-				regs [(int) I386Register.EIP].SetValue (new_esp, new_eip);
+				regs [(int) X86_Register.RIP].SetValue (new_esp, new_eip);
 				new_esp -= addr_size;
 
-				regs [(int) I386Register.ESP].SetValue (new_esp);
+				regs [(int) X86_Register.RSP].SetValue (new_esp);
 
-				return CreateFrame (frame.Thread, new_eip, new_esp, new_ebp, regs);
+				return CreateFrame (frame.Thread, memory, new_eip, new_esp,
+						    new_ebp, regs, true);
 			}
 
 			// mov %ebp, %esp
@@ -516,10 +343,10 @@ namespace Mono.Debugger.Backends
 		{
 			Registers regs = new Registers (old_regs);
 
-			regs [(int) I386Register.EBP].Valid = true;
-			regs [(int) I386Register.EBX].Valid = true;
-			regs [(int) I386Register.ESI].Valid = true;
-			regs [(int) I386Register.EDI].Valid = true;
+			regs [(int) X86_Register.RBP].Valid = true;
+			regs [(int) X86_Register.RBX].Valid = true;
+			regs [(int) X86_Register.RSI].Valid = true;
+			regs [(int) X86_Register.RDI].Valid = true;
 
 			return regs;
 		}
@@ -570,16 +397,16 @@ namespace Mono.Debugger.Backends
 			TargetAddress ebp = memory.ReadAddress (esp + 0x38);
 			TargetAddress eip = memory.ReadAddress (esp + 0x3c);
 
-			regs [(int)I386Register.EBX].SetValue (esp + 0x2c, ebx);
-			regs [(int)I386Register.ESI].SetValue (esp + 0x30, esi);
-			regs [(int)I386Register.EDI].SetValue (esp + 0x34, edi);
-			regs [(int)I386Register.EBP].SetValue (esp + 0x38, ebp);
-			regs [(int)I386Register.EIP].SetValue (esp + 0x3c, eip);
+			regs [(int)X86_Register.RBX].SetValue (esp + 0x2c, ebx);
+			regs [(int)X86_Register.RSI].SetValue (esp + 0x30, esi);
+			regs [(int)X86_Register.RDI].SetValue (esp + 0x34, edi);
+			regs [(int)X86_Register.RBP].SetValue (esp + 0x38, ebp);
+			regs [(int)X86_Register.RIP].SetValue (esp + 0x3c, eip);
 
 			esp += 0x40;
-			regs [(int)I386Register.ESP].SetValue (esp.Address);
+			regs [(int)X86_Register.RSP].SetValue (esp.Address);
 
-			return CreateFrame (frame.Thread, eip, esp, ebp, regs);
+			return CreateFrame (frame.Thread, memory, eip, esp, ebp, regs, true);
 		}
 
 		StackFrame try_syscall_trampoline (StackFrame frame, TargetMemoryAccess memory)
@@ -599,10 +426,11 @@ namespace Mono.Debugger.Backends
 			TargetAddress new_esp = esp + 4;
 			TargetAddress new_ebp = frame.FrameAddress;
 
-			regs [(int)I386Register.EIP].SetValue (esp, new_eip);
-			regs [(int)I386Register.EBP].SetValue (new_ebp);
+			regs [(int)X86_Register.RIP].SetValue (esp, new_eip);
+			regs [(int)X86_Register.RBP].SetValue (new_ebp);
 
-			return CreateFrame (frame.Thread, new_eip, new_esp, new_ebp, regs);
+			return CreateFrame (frame.Thread, memory, new_eip, new_esp,
+					    new_ebp, regs, true);
 		}
 
 		StackFrame do_hacks (StackFrame frame, TargetMemoryAccess memory)
@@ -630,12 +458,16 @@ namespace Mono.Debugger.Backends
 		internal override StackFrame UnwindStack (StackFrame frame, TargetMemoryAccess memory,
 							  byte[] code, int offset)
 		{
-			if ((code != null) && (code.Length > 3))
-				return read_prologue (frame, memory, code, offset);
+			StackFrame new_frame;
+			if ((code != null) && (code.Length > 3)) {
+				new_frame = read_prologue (frame, memory, code, offset);
+				if (new_frame != null)
+					return new_frame;
+			}
 
 			TargetAddress ebp = frame.FrameAddress;
 
-			StackFrame new_frame = do_hacks (frame, memory);
+			new_frame = do_hacks (frame, memory);
 			if (new_frame != null)
 				return new_frame;
 
@@ -644,17 +476,18 @@ namespace Mono.Debugger.Backends
 			Registers regs = new Registers (this);
 
 			TargetAddress new_ebp = memory.ReadAddress (ebp);
-			regs [(int) I386Register.EBP].SetValue (ebp, new_ebp);
+			regs [(int) X86_Register.RBP].SetValue (ebp, new_ebp);
 
 			TargetAddress new_eip = memory.ReadAddress (ebp + addr_size);
-			regs [(int) I386Register.EIP].SetValue (ebp + addr_size, new_eip);
+			regs [(int) X86_Register.RIP].SetValue (ebp + addr_size, new_eip);
 
 			TargetAddress new_esp = ebp + 2 * addr_size;
-			regs [(int) I386Register.ESP].SetValue (ebp, new_esp);
+			regs [(int) X86_Register.RSP].SetValue (ebp, new_esp);
 
 			ebp -= addr_size;
 
-			return CreateFrame (frame.Thread, new_eip, new_esp, new_ebp, regs);
+			return CreateFrame (frame.Thread, memory, new_eip, new_esp,
+					    new_ebp, regs, true);
 		}
 
 		internal override StackFrame TrySpecialUnwind (StackFrame last_frame,
@@ -667,21 +500,21 @@ namespace Mono.Debugger.Backends
 		{
 			Registers regs = inferior.GetRegisters ();
 			TargetAddress esp = new TargetAddress (
-				AddressDomain, regs [(int) I386Register.ESP].GetValue ());
+				inferior.AddressDomain, regs [(int) X86_Register.RSP].GetValue ());
 			TargetAddress eip = inferior.ReadAddress (esp);
 			esp += TargetAddressSize;
 
-			regs [(int) I386Register.EIP].SetValue (eip);
-			regs [(int) I386Register.ESP].SetValue (esp);
-			regs [(int) I386Register.EAX].SetValue (TargetAddress.Null);
+			regs [(int) X86_Register.RIP].SetValue (eip);
+			regs [(int) X86_Register.RSP].SetValue (esp);
+			regs [(int) X86_Register.RAX].SetValue (TargetAddress.Null);
 
 			inferior.SetRegisters (regs);
 		}
 
-		protected override TargetAddress AdjustReturnAddress (Thread thread,
+		protected override TargetAddress AdjustReturnAddress (TargetMemoryAccess target,
 								      TargetAddress address)
 		{
-			TargetBinaryReader reader = thread.ReadMemory (address-7, 7).GetReader ();
+			TargetBinaryReader reader = target.ReadMemory (address-7, 7).GetReader ();
 
 			byte[] code = reader.ReadBuffer (7);
 			if (code [2] == 0xe8)
@@ -703,49 +536,48 @@ namespace Mono.Debugger.Backends
 			return address;
 		}
 
-		internal override StackFrame CreateFrame (Thread thread, Registers regs)
+		internal override StackFrame CreateFrame (Thread thread, TargetMemoryAccess memory,
+							  Registers regs, bool adjust_retaddr)
 		{
 			TargetAddress address = new TargetAddress (
-				AddressDomain, regs [(int) I386Register.EIP].GetValue ());
+				memory.AddressDomain, regs [(int) X86_Register.RIP].GetValue ());
 			TargetAddress stack_pointer = new TargetAddress (
-				AddressDomain, regs [(int) I386Register.ESP].GetValue ());
+				memory.AddressDomain, regs [(int) X86_Register.RSP].GetValue ());
 			TargetAddress frame_pointer = new TargetAddress (
-				AddressDomain, regs [(int) I386Register.EBP].GetValue ());
+				memory.AddressDomain, regs [(int) X86_Register.RBP].GetValue ());
 
-			return CreateFrame (thread, address, stack_pointer, frame_pointer, regs);
+			return CreateFrame (thread, memory, address, stack_pointer, frame_pointer,
+					    regs, adjust_retaddr);
 		}
 
-		internal override StackFrame GetLMF (Thread thread)
+		internal override StackFrame GetLMF (ThreadServant thread, TargetMemoryAccess memory)
 		{
-			TargetAddress lmf = thread.ReadAddress (thread.LMFAddress);
+			TargetAddress lmf = memory.ReadAddress (thread.LMFAddress);
 
-			TargetBinaryReader reader = thread.ReadMemory (lmf, 32).GetReader ();
+			TargetBinaryReader reader = memory.ReadMemory (lmf, 36).GetReader ();
 
-			reader.Position = 12;
+			reader.Position = 16;
 
 			TargetAddress ebx = reader.ReadTargetAddress ();
 			TargetAddress edi = reader.ReadTargetAddress ();
 			TargetAddress esi = reader.ReadTargetAddress ();
 			TargetAddress ebp = reader.ReadTargetAddress ();
+			TargetAddress eip = reader.ReadTargetAddress ();
 
 			Registers regs = new Registers (this);
-			regs [(int) I386Register.EBX].SetValue (lmf + 12, ebx);
-			regs [(int) I386Register.EDI].SetValue (lmf + 16, edi);
-			regs [(int) I386Register.ESI].SetValue (lmf + 20, esi);
-			regs [(int) I386Register.EBP].SetValue (lmf + 24, ebp);
+			regs [(int) X86_Register.RBX].SetValue (lmf + 16, ebx);
+			regs [(int) X86_Register.RDI].SetValue (lmf + 20, edi);
+			regs [(int) X86_Register.RSI].SetValue (lmf + 24, esi);
+			regs [(int) X86_Register.RBP].SetValue (lmf + 28, ebp);
+			regs [(int) X86_Register.RIP].SetValue (lmf + 32, eip);
 
-			TargetAddress new_eip = thread.ReadAddress (ebp + 4);
-			regs [(int) I386Register.EIP].SetValue (ebp + 4, new_eip);
-
-			TargetAddress new_ebp = thread.ReadAddress (ebp);
-			regs [(int) I386Register.EBP].SetValue (ebp, new_ebp);
+			TargetAddress new_ebp = memory.ReadAddress (ebp);
+			regs [(int) X86_Register.RBP].SetValue (ebp, new_ebp);
 
 			TargetAddress new_esp = ebp + 8;
-			regs [(int) I386Register.ESP].SetValue (ebp, new_esp);
+			regs [(int) X86_Register.RSP].SetValue (ebp, new_esp);
 
-			ebp -= 4;
-
-			return CreateFrame (thread, new_eip, new_esp, new_ebp, regs);
+			return CreateFrame (thread.Client, memory, eip, new_esp, new_ebp, regs, true);
 		}
 	}
 }
