@@ -2,6 +2,7 @@ using System;
 using System.IO;
 using System.Text;
 using System.Collections;
+using System.Collections.Generic;
 using System.Diagnostics;
 using System.Runtime.InteropServices;
 
@@ -290,7 +291,8 @@ namespace Mono.Debugger.Backend
 		{
 			SourceFile file = (SourceFile) source_file_hash [filename];
 			if (file == null) {
-				file = new SourceFile (module, filename);
+				file = new DwarfSourceFile (
+					bfd.BfdContainer.Process.Session, module, filename);
 				source_file_hash.Add (filename, file);
 			}
 			return file;
@@ -313,6 +315,23 @@ namespace Mono.Debugger.Backend
 			types_initialized = true;
 		}
 
+		protected class DwarfSourceFile : SourceFile
+		{
+			public DwarfSourceFile (DebuggerSession session, Module module,
+						string filename)
+				: base (session, module, filename)
+			{ }
+
+			public override bool IsAutoGenerated {
+				get { return false; }
+			}
+
+			public override bool CheckModified ()
+			{
+				return false;
+			}
+		}
+
 		protected class CompileUnitBlock
 		{
 			public readonly DwarfReader dwarf;
@@ -1056,16 +1075,12 @@ namespace Mono.Debugger.Backend
 				public int start_file;
 				public int start_line, end_line;
 
-				bool creating;
-				DwarfBinaryReader reader;
 				int const_add_pc_range;
 
 				public StatementMachine (LineNumberEngine engine, long offset,
 							 long end_offset)
 				{
 					this.engine = engine;
-					this.reader = this.engine.reader;
-					this.creating = true;
 					this.start_offset = offset;
 					this.end_offset = end_offset;
 					this.st_address = 0;
@@ -1407,6 +1422,18 @@ namespace Mono.Debugger.Backend
 				return do_lookup (address, 0, addresses.Length-1);
 			}
 
+			public override bool HasMethodBounds {
+				get { return false; }
+			}
+
+			public override TargetAddress MethodStartAddress {
+				get { throw new InvalidOperationException (); }
+			}
+
+			public override TargetAddress MethodEndAddress {
+				get { throw new InvalidOperationException (); }
+			}
+
 			public override void DumpLineNumbers ()
 			{
 				Console.WriteLine ("--------");
@@ -2519,7 +2546,7 @@ namespace Mono.Debugger.Backend
 				get { return -1; }
 			}
 
-			public override TargetClassType GetDeclaringType (Thread target)
+			public override TargetStructType GetDeclaringType (Thread target)
 			{
 				return null;
 			}
@@ -2543,6 +2570,14 @@ namespace Mono.Debugger.Backend
 				return subprog.Locals;
 			}
 
+			public override bool IsWrapper {
+				get { return false; }
+			}
+
+			public override bool IsCompilerGenerated {
+				get { return false; }
+			}
+
 			public override bool HasSource {
 				get {
 					read_line_numbers ();
@@ -2613,17 +2648,6 @@ namespace Mono.Debugger.Backend
 
 				read_line_numbers ();
 
-#if FIXME
-				if ((lines != null) && (lines.Length > 2)) {
-					LineNumber start = lines [1];
-					LineNumber end = lines [lines.Length - 1];
-
-					SetMethodBounds (
-						subprog.dwarf.GetAddress (start.Offset),
-						subprog.dwarf.GetAddress (end.Offset));
-				}
-#endif
-
 				return true;
 			}
 
@@ -2752,7 +2776,6 @@ namespace Mono.Debugger.Backend
 			DieSubprogram subprog;
 			byte[] location_block;
 			long loclist_offset;
-			bool has_loclist;
 			bool is_byref;
 
 			public DwarfLocation (DieSubprogram subprog, Attribute attribute, bool is_byref)
@@ -2772,7 +2795,6 @@ namespace Mono.Debugger.Backend
 				case DwarfForm.data4:
 				case DwarfForm.data8:
 					loclist_offset = (long) attribute.Data;
-					has_loclist = true;
 					break;
 				default:
 					throw new InternalError  ();
@@ -2820,6 +2842,14 @@ namespace Mono.Debugger.Backend
 					reg = locreader.ReadLeb128 () + 3;
 					off = locreader.ReadSLeb128 ();
 					is_regoffset = true;
+				} else if (opcode == 0x03) { // DW_OP_addr
+					TargetAddress addr = new TargetAddress (
+						memory.AddressDomain, locreader.ReadAddress ());
+					TargetLocation aloc = new AbsoluteTargetLocation (addr);
+					if (is_byref)
+						return new DereferencedTargetLocation (aloc);
+					else
+						return aloc;
 				} else {
 					Console.WriteLine ("UNKNOWN OPCODE: {0:x}", opcode);
 					return null;
@@ -2878,15 +2908,12 @@ namespace Mono.Debugger.Backend
 
 		protected class DwarfTargetVariable : TargetVariable
 		{
-			readonly DieSubprogram subprog;
 			readonly string name;
 			readonly TargetType type;
 			readonly DwarfLocation location;
-			// int offset;
 
 			protected DwarfTargetVariable (DieSubprogram subprog, string name, TargetType type)
 			{
-				this.subprog = subprog;
 				this.name = name;
 				this.type = type;
 			}
@@ -2921,16 +2948,14 @@ namespace Mono.Debugger.Backend
 				return true;
 			}
 
-			public override TargetObject GetObject (StackFrame frame)
+			internal override TargetObject GetObject (StackFrame frame,
+								  TargetMemoryAccess target)
 			{
-				return (TargetObject) frame.Thread.ThreadServant.DoTargetAccess (
-					delegate (TargetMemoryAccess memory) {
-						TargetLocation loc = location.GetLocation (frame, memory);
-						if (loc == null)
-							return null;
+				TargetLocation loc = location.GetLocation (frame, target);
+				if (loc == null)
+					return null;
 
-						return type.GetObject (memory, loc);
-				});
+				return type.GetObject (target, loc);
 			}
 
 			public override string PrintLocation (StackFrame frame)
@@ -3219,19 +3244,23 @@ namespace Mono.Debugger.Backend
 					}
 
 					ref_type = reference.ResolveType ();
+					if (ref_type == null)
+						return null;
 				}
-				if (ref_type == null)
-					return null;
 
 				TargetFundamentalType fundamental = ref_type as TargetFundamentalType;
 				if ((fundamental != null) && (fundamental.Name == "char"))
 					return new NativeStringType (language, byte_size);
 
+				TargetFunctionType func = ref_type as TargetFunctionType;
+				if (func != null)
+					return new NativeFunctionPointer (language, func);
+
 				string name;
 				if (Name != null)
 					name = Name;
 				else
-					name = String.Format ("{0}*", ref_type.Name);
+					name = NativePointerType.MakePointerName (ref_type);
 
 				return new NativePointerType (language, name, ref_type, byte_size);
 			}
@@ -3239,8 +3268,15 @@ namespace Mono.Debugger.Backend
 
 		protected class DieSubrangeType : Die
 		{
-			int upper_bound;
-			int lower_bound;
+			public int? UpperBound {
+				get;
+				private set;
+			}
+
+			public int? LowerBound {
+				get;
+				private set;
+			}
 
 			public DieSubrangeType (DwarfBinaryReader reader, CompilationUnit comp_unit,
 						AbbrevEntry abbrev)
@@ -3251,31 +3287,19 @@ namespace Mono.Debugger.Backend
 			{
 				switch (attribute.DwarfAttribute) {
 				case DwarfAttribute.upper_bound:
-					upper_bound = (int) (long) attribute.Data;
+					UpperBound = (int) (long) attribute.Data;
 					break;
 
 				case DwarfAttribute.lower_bound:
-					lower_bound = (int) (long) attribute.Data;
+					LowerBound = (int) (long) attribute.Data;
 					break;
 
 				case DwarfAttribute.count:
-					lower_bound = 0;
-				  	upper_bound = ((int) (long) attribute.Data) - 1;
+					LowerBound = 0;
+				  	UpperBound = (int) (long) attribute.Data;
 					break;
 				}
 			}
-
-		  	public int UpperBound {
-				get {
-					return upper_bound;
-				}
-			}
-
-			public int LowerBound {
-				get {
-					return lower_bound;
-				}
-			}
 		}
 
 		protected class DieArrayType : DieType
@@ -3354,21 +3378,35 @@ namespace Mono.Debugger.Backend
 				else
 					name = String.Format ("{0} []", ref_type.Name);
 
-				/* XXX for now just find the first
-				 * Subrange child and use that for the
-				 * array dimensions.  This should
-				 * really support multidimensional
-				 * arrays */
-				DieSubrangeType subrange = null;
-				foreach (Die d in Children) {
-					subrange = d as DieSubrangeType;
-					if (subrange == null) continue;
-					break;
+				List<DieSubrangeType> list = new List<DieSubrangeType> ();
+
+				foreach (Die die in Children) {
+					DieSubrangeType subrange = die as DieSubrangeType;
+					if (subrange != null)
+						list.Add (subrange);
+				}
+
+				TargetArrayBounds bounds;
+				if (list.Count == 0)
+					bounds = TargetArrayBounds.MakeUnboundArray ();
+				else if ((list.Count == 1) && (list [0].UpperBound == null))
+					bounds = TargetArrayBounds.MakeUnboundArray ();
+				else if ((list.Count == 1) && ((list [0].LowerBound ?? 0) == 0))
+					bounds = TargetArrayBounds.MakeSimpleArray ((int) list [0].UpperBound + 1);
+				else {
+					int[] lower = new int [list.Count];
+					int[] upper = new int [list.Count];
+
+					for (int i = 0; i < list.Count; i++) {
+						lower [i] = list [i].LowerBound ?? 0;
+						upper [i] = (int) list [i].UpperBound;
+					}
+
+					bounds = TargetArrayBounds.MakeMultiArray (lower, upper);
 				}
 
 				return new NativeArrayType (
-					language, name, ref_type,
-					subrange.LowerBound, subrange.UpperBound, byte_size);
+					language, name, ref_type, bounds, byte_size);
 			}
 		}
 
@@ -3540,7 +3578,7 @@ namespace Mono.Debugger.Backend
 
 			protected override void PopulateType ()
 			{
-				type.TargetType = reference.ResolveType ();
+				type.SetTargetType (reference.ResolveType ());
 			}
 		}
 
@@ -3690,6 +3728,7 @@ namespace Mono.Debugger.Backend
 			long type_offset;
 			bool prototyped;
 			DieType return_type;
+			NativeFunctionType function_type;
 
 			public DieSubroutineType (DwarfBinaryReader reader, CompilationUnit comp_unit,
 						  long offset, AbbrevEntry abbrev)
@@ -3736,32 +3775,34 @@ namespace Mono.Debugger.Backend
 						return null;
 				}
 
+				function_type = new NativeFunctionType (language);
+				return function_type;
+			}
+
+			protected override void PopulateType ()
+			{
 				TargetType ret_type = null;
 				if (return_type != null)
 					ret_type = return_type.ResolveType ();
 				if (ret_type == null)
 					ret_type = (TargetType) language.VoidType;
 
-				TargetType[] param_types = new TargetType [0];
-				NativeFunctionType func_type = new NativeFunctionType (
-					language, "test", ret_type, param_types);
-				return func_type;
-			}
+				List<TargetType> param_list = new List<TargetType> ();
 
-			protected override void PopulateType ()
-			{
-				if (!abbrev.HasChildren)
-					return;
+				if (abbrev.HasChildren) {
+					foreach (Die child in Children) {
+						DieFormalParameter formal = child as DieFormalParameter;
+						if (formal == null) {
+							Console.WriteLine ("UNKNOWN DIE IN PROTOTYPE: {0}",
+									   child);
+							continue;
+						}
 
-				ArrayList args = new ArrayList ();
-
-				foreach (Die child in Children) {
-					DieFormalParameter formal = child as DieFormalParameter;
-					if (formal == null)
-						continue;
-
-					args.Add (formal);
+						param_list.Add (formal.Type);
+					}
 				}
+
+				function_type.SetPrototype (ret_type, param_list.ToArray ());
 			}
 		}
 
@@ -3812,7 +3853,6 @@ namespace Mono.Debugger.Backend
 						  bool is_local)
 				: base (reader, comp_unit, abbrev)
 			{
-				this.target_info = reader.TargetMemoryInfo;
 				this.subprog = subprog;
 
 				if (subprog != null) {
@@ -3838,29 +3878,46 @@ namespace Mono.Debugger.Backend
 
 			Attribute location_attr;
 			TargetVariable variable;
-			TargetMemoryInfo target_info;
 			DieSubprogram subprog;
+			TargetType type;
 			bool resolved;
 
-			protected bool DoResolve ()
+			protected bool DoResolveType ()
 			{
-				if ((TypeOffset == 0) || (Name == null) || (subprog == null) ||
-				    (location_attr == null))
-					return false;
+				if (type != null)
+					return true;
 
+				if (TypeOffset == 0)
+					return false;
 				DieType reference = comp_unit.GetType (TypeOffset);
 				if (reference == null)
 					return false;
 
-				TargetType type = reference.ResolveType ();
-				if (type == null)
+				type = reference.ResolveType ();
+				return type != null;
+			}
+
+			protected bool DoResolve ()
+			{
+				if (!DoResolveType ())
 					return false;
 
-				DwarfLocation location = new DwarfLocation (subprog, location_attr, type.IsByRef);
+				if ((Name == null) || (subprog == null) || (location_attr == null))
+					return false;
+
+				DwarfLocation location = new DwarfLocation (
+					subprog, location_attr, type.IsByRef);
 				variable = new DwarfTargetVariable (subprog, Name, type, location);
 				return true;
 			}
 
+			public TargetType Type {
+				get {
+					DoResolveType ();
+					return type;
+				}
+			}
+
 			public TargetVariable Variable {
 				get {
 					if (!resolved) {
