@@ -161,10 +161,11 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 	*((guint32 *) (code+16)) = method_argument1 & 0xffffffff;
 	*((guint32 *) (code+21)) = call_disp - 25;
 
-	server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
+	result = server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_EAX (arch->current_regs) = -1;
 	INFERIOR_REG_ESP (arch->current_regs) = INFERIOR_REG_EIP (arch->current_regs) = new_esp;
 
 	g_ptr_array_add (arch->callback_stack, cdata);
@@ -231,6 +232,7 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_EAX (arch->current_regs) = -1;
 	INFERIOR_REG_ESP (arch->current_regs) = INFERIOR_REG_EIP (arch->current_regs) = new_esp;
 
 	g_ptr_array_add (arch->callback_stack, cdata);
@@ -288,11 +290,12 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 		cdata->data_size = data_size;
 	}
 
-	server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
+	result = server_ptrace_write_memory (handle, (guint32) new_esp, size, code);
 	g_free (code);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_EAX (arch->current_regs) = -1;
 	INFERIOR_REG_EIP (arch->current_regs) = method_address;
 	INFERIOR_REG_ESP (arch->current_regs) = new_esp;
 
@@ -305,6 +308,70 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 	return server_ptrace_continue (handle);
 }
 
+static ServerCommandError
+server_ptrace_call_method_3 (ServerHandle *handle, guint64 method_address,
+			     guint64 method_argument, guint64 address_argument,
+			     guint32 blob_size, gconstpointer blob_data,
+			     guint64 callback_argument)
+{
+	ServerCommandError result = COMMAND_ERROR_NONE;
+	ArchInfo *arch = handle->arch;
+	CallbackData *cdata;
+	guint32 new_esp;
+	int i;
+
+	static guint8 static_code[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+					0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+					0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+					0xcc };
+	int static_size = sizeof (static_code);
+	int size = static_size + blob_size;
+	guint8 *code = g_malloc0 (size);
+	guint32 *ptr = (guint32 *) (code + static_size + blob_size);
+	guint32 effective_address;
+	guint32 blob_start;
+	memcpy (code, static_code, static_size);
+	memcpy (code + static_size, blob_data, blob_size);
+
+	new_esp = INFERIOR_REG_ESP (arch->current_regs) - size;
+
+	blob_start = new_esp + static_size;
+	effective_address = address_argument ? address_argument : blob_start;
+
+	*((guint32 *) code) = new_esp + static_size - 1;
+	*((guint32 *) (code+4)) = method_argument;
+	*((guint32 *) (code+12)) = effective_address;
+	*((guint32 *) (code+20)) = new_esp;
+
+	cdata = g_new0 (CallbackData, 1);
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_esp + static_size;
+	cdata->stack_pointer = new_esp + 4;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
+	handle->inferior->last_signal = 0;
+
+	result = server_ptrace_write_memory (handle, (unsigned long) new_esp, size, code);
+	g_free (code);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	INFERIOR_REG_ORIG_EAX (arch->current_regs) = -1;
+	INFERIOR_REG_EIP (arch->current_regs) = method_address;
+	INFERIOR_REG_ESP (arch->current_regs) = new_esp;
+	INFERIOR_REG_EBP (arch->current_regs) = 0;
+
+	g_ptr_array_add (arch->callback_stack, cdata);
+
+	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	return server_ptrace_continue (handle);
+
+}
+
 static ServerCommandError
 server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 				  guint64 method_argument, guint32 num_params,
@@ -363,6 +430,7 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_EAX (arch->current_regs) = -1;
 	INFERIOR_REG_EIP (arch->current_regs) = invoke_method;
 	INFERIOR_REG_ESP (arch->current_regs) = new_esp;
 	INFERIOR_REG_EBP (arch->current_regs) = 0;
@@ -444,6 +512,9 @@ x86_arch_child_stopped (ServerHandle *handle, int stopsig,
 		guint32 addr = (guint32) INFERIOR_REG_ESP (arch->current_regs) + 4;
 		guint64 data [3];
 
+		if (stopsig != SIGTRAP)
+			return STOP_ACTION_STOPPED;
+
 		if (server_ptrace_read_memory (handle, addr, 24, &data))
 			return STOP_ACTION_STOPPED;
 
@@ -670,6 +741,29 @@ server_ptrace_pop_registers (ServerHandle *handle)
 	return COMMAND_ERROR_NONE;
 }
 
+static void
+server_ptrace_get_registers_from_core_file (guint64 *values, const guint8 *buffer)
+{
+	INFERIOR_REGS_TYPE regs = * (INFERIOR_REGS_TYPE *) buffer;
+
+	values [DEBUGGER_REG_RBX] = (guint32) INFERIOR_REG_EBX (regs);
+	values [DEBUGGER_REG_RCX] = (guint32) INFERIOR_REG_ECX (regs);
+	values [DEBUGGER_REG_RDX] = (guint32) INFERIOR_REG_EDX (regs);
+	values [DEBUGGER_REG_RSI] = (guint32) INFERIOR_REG_ESI (regs);
+	values [DEBUGGER_REG_RDI] = (guint32) INFERIOR_REG_EDI (regs);
+	values [DEBUGGER_REG_RBP] = (guint32) INFERIOR_REG_EBP (regs);
+	values [DEBUGGER_REG_RAX] = (guint32) INFERIOR_REG_EAX (regs);
+	values [DEBUGGER_REG_DS] = (guint32) INFERIOR_REG_DS (regs);
+	values [DEBUGGER_REG_ES] = (guint32) INFERIOR_REG_ES (regs);
+	values [DEBUGGER_REG_FS] = (guint32) INFERIOR_REG_FS (regs);
+	values [DEBUGGER_REG_GS] = (guint32) INFERIOR_REG_GS (regs);
+	values [DEBUGGER_REG_RIP] = (guint32) INFERIOR_REG_EIP (regs);
+	values [DEBUGGER_REG_CS] = (guint32) INFERIOR_REG_CS (regs);
+	values [DEBUGGER_REG_EFLAGS] = (guint32) INFERIOR_REG_EFLAGS (regs);
+	values [DEBUGGER_REG_RSP] = (guint32) INFERIOR_REG_ESP (regs);
+	values [DEBUGGER_REG_SS] = (guint32) INFERIOR_REG_SS (regs);
+}
+
 static int
 find_breakpoint_table_slot (MonoRuntimeInfo *runtime)
 {
@@ -1103,6 +1197,7 @@ server_ptrace_execute_instruction (ServerHandle *handle, const guint8 *instructi
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_EAX (handle->arch->current_regs) = -1;
 	INFERIOR_REG_EIP (handle->arch->current_regs) = code_address;
 
 	result = _server_ptrace_set_registers (handle->inferior, &handle->arch->current_regs);
