@@ -86,15 +86,17 @@ typedef struct InferiorHandle InferiorHandle;
 typedef struct ServerHandle ServerHandle;
 typedef struct ArchInfo ArchInfo;
 
+typedef struct IOThreadData IOThreadData;
+
 /* C# delegates. */
-typedef void (*ChildOutputFunc) (const char *output);
+typedef void (*ChildOutputFunc) (gboolean is_stderr, const char *output);
 
 typedef struct {
 	int sigkill;
 	int sigstop;
 	int sigint;
 	int sigchld;
-
+	int kernel_sigrtmin;
 	int mono_thread_abort;
 } SignalInfo;
 
@@ -129,13 +131,15 @@ struct InferiorVTable {
 	void                  (* set_runtime_info)    (ServerHandle       *handle,
 						       MonoRuntimeInfo    *mono_runtime_info);
 
+	void                  (* io_thread_main)      (IOThreadData       *io_data,
+						       ChildOutputFunc     func);
+
 	ServerCommandError    (* spawn)               (ServerHandle       *handle,
 						       const gchar        *working_directory,
 						       const gchar       **argv,
 						       const gchar       **envp,
 						       gint               *child_pid,
-						       ChildOutputFunc     stdout_handler,
-						       ChildOutputFunc     stderr_handler,
+						       IOThreadData      **io_data,
 						       gchar             **error);
 
 	ServerCommandError    (* attach)              (ServerHandle       *handle,
@@ -246,6 +250,14 @@ struct InferiorVTable {
 						       gconstpointer     data_buffer,
 						       guint64           callback_argument);
 
+	ServerCommandError    (* call_method_3)       (ServerHandle     *handle,
+						       guint64           method_address,
+						       guint64           method_argument,
+						       guint64           address_argument,
+						       guint32           blob_size,
+						       gconstpointer     blob_data,
+						       guint64           callback_argument);
+
 	ServerCommandError    (* call_method_invoke)  (ServerHandle     *handle,
 						       guint64           invoke_method,
 						       guint64           method_argument,
@@ -369,6 +381,9 @@ struct InferiorVTable {
 						       guint64            stack_pointer,
 						       gboolean           exact_match,
 						       guint64           *registers);
+
+	void         (* get_registers_from_core_file) (guint64           *values,
+						       const guint8      *buffer);
 };
 
 /*
@@ -394,14 +409,17 @@ ServerCommandError
 mono_debugger_server_initialize_thread    (ServerHandle       *handle,
 					   guint32             pid);
 
+void
+mono_debugger_server_io_thread_main       (IOThreadData       *io_data,
+					   ChildOutputFunc     func);
+
 ServerCommandError
 mono_debugger_server_spawn                (ServerHandle       *handle,
 					   const gchar        *working_directory,
 					   const gchar       **argv,
 					   const gchar       **envp,
 					   gint               *child_pid,
-					   ChildOutputFunc     stdout_handler,
-					   ChildOutputFunc     stderr_handler,
+					   IOThreadData      **io_data,
 					   gchar             **error);
 
 ServerCommandError
@@ -486,6 +504,15 @@ mono_debugger_server_call_method_2        (ServerHandle       *handle,
 					   gconstpointer       data_buffer,
 					   guint64             callback_argument);
 
+ServerCommandError
+mono_debugger_server_call_method_3        (ServerHandle       *handle,
+					   guint64             method_address,
+					   guint64             method_argument,
+					   guint64             address_argument,
+					   guint32             blob_size,
+					   gconstpointer       blob_data,
+					   guint64             callback_argument);
+
 ServerCommandError
 mono_debugger_server_call_method_invoke   (ServerHandle       *handle,
 					   guint64             invoke_method,
@@ -606,6 +633,15 @@ mono_debugger_server_initialize_mono_runtime (guint32 address_size,
 void
 mono_debugger_server_finalize_mono_runtime (MonoRuntimeInfo *runtime);
 
+void
+mono_debugger_server_initialize_code_buffer (MonoRuntimeInfo *runtime,
+					     guint64 executable_code_buffer,
+					     guint32 executable_code_buffer_size);
+
+void
+mono_debugger_server_get_registers_from_core_file (guint64 *values,
+						   const guint8 *buffer);
+
 /* POSIX semaphores */
 
 void mono_debugger_server_sem_init (void);
