@@ -40,6 +40,9 @@ server_ptrace_stop_and_wait (ServerHandle *handle, guint32 *status);
 static ServerCommandError
 _server_ptrace_setup_inferior (ServerHandle *handle);
 
+static void
+_server_ptrace_finalize_inferior (ServerHandle *handle);
+
 static ServerCommandError
 server_ptrace_get_signal_info (ServerHandle *handle, SignalInfo **sinfo);
 
