@@ -188,6 +188,9 @@ x86_arch_child_stopped (ServerHandle *handle, int stopsig,
 
 	if (handle->mono_runtime &&
 	    (INFERIOR_REG_RIP (arch->current_regs) - 1 == handle->mono_runtime->notification_address)) {
+		if (stopsig != SIGTRAP)
+			return STOP_ACTION_STOPPED;
+
 		*callback_arg = INFERIOR_REG_RDI (arch->current_regs);
 		*retval = INFERIOR_REG_RSI (arch->current_regs);
 		*retval2 = INFERIOR_REG_RDX (arch->current_regs);
@@ -444,6 +447,40 @@ server_ptrace_pop_registers (ServerHandle *handle)
 	return COMMAND_ERROR_NONE;
 }
 
+static void
+server_ptrace_get_registers_from_core_file (guint64 *values, const guint8 *buffer)
+{
+	INFERIOR_REGS_TYPE regs = * (INFERIOR_REGS_TYPE *) buffer;
+
+	values [DEBUGGER_REG_R15] = (guint64) INFERIOR_REG_R15 (regs);
+	values [DEBUGGER_REG_R14] = (guint64) INFERIOR_REG_R14 (regs);
+	values [DEBUGGER_REG_R13] = (guint64) INFERIOR_REG_R13 (regs);
+	values [DEBUGGER_REG_R12] = (guint64) INFERIOR_REG_R12 (regs);
+	values [DEBUGGER_REG_RBP] = (guint64) INFERIOR_REG_RBP (regs);
+	values [DEBUGGER_REG_RBX] = (guint64) INFERIOR_REG_RBX (regs);
+	values [DEBUGGER_REG_R11] = (guint64) INFERIOR_REG_R11 (regs);
+	values [DEBUGGER_REG_R10] = (guint64) INFERIOR_REG_R10 (regs);
+	values [DEBUGGER_REG_R9] = (guint64) INFERIOR_REG_R9 (regs);
+	values [DEBUGGER_REG_R8] = (guint64) INFERIOR_REG_R8 (regs);
+	values [DEBUGGER_REG_RAX] = (guint64) INFERIOR_REG_RAX (regs);
+	values [DEBUGGER_REG_RCX] = (guint64) INFERIOR_REG_RCX (regs);
+	values [DEBUGGER_REG_RDX] = (guint64) INFERIOR_REG_RDX (regs);
+	values [DEBUGGER_REG_RSI] = (guint64) INFERIOR_REG_RSI (regs);
+	values [DEBUGGER_REG_RDI] = (guint64) INFERIOR_REG_RDI (regs);
+	values [DEBUGGER_REG_ORIG_RAX] = (guint64) INFERIOR_REG_ORIG_RAX (regs);
+	values [DEBUGGER_REG_RIP] = (guint64) INFERIOR_REG_RIP (regs);
+	values [DEBUGGER_REG_CS] = (guint64) INFERIOR_REG_CS (regs);
+	values [DEBUGGER_REG_EFLAGS] = (guint64) INFERIOR_REG_EFLAGS (regs);
+	values [DEBUGGER_REG_RSP] = (guint64) INFERIOR_REG_RSP (regs);
+	values [DEBUGGER_REG_SS] = (guint64) INFERIOR_REG_SS (regs);
+	values [DEBUGGER_REG_FS_BASE] = (guint64) INFERIOR_REG_FS_BASE (regs);
+	values [DEBUGGER_REG_GS_BASE] = (guint64) INFERIOR_REG_GS_BASE (regs);
+	values [DEBUGGER_REG_DS] = (guint64) INFERIOR_REG_DS (regs);
+	values [DEBUGGER_REG_ES] = (guint64) INFERIOR_REG_ES (regs);
+	values [DEBUGGER_REG_FS] = (guint64) INFERIOR_REG_FS (regs);
+	values [DEBUGGER_REG_GS] = (guint64) INFERIOR_REG_GS (regs);
+}
+
 static int
 find_breakpoint_table_slot (MonoRuntimeInfo *runtime)
 {
@@ -859,10 +896,11 @@ server_ptrace_call_method (ServerHandle *handle, guint64 method_address,
 	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
-	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
+	result = server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_RAX (arch->current_regs) = -1;
 	INFERIOR_REG_RIP (arch->current_regs) = method_address;
 	INFERIOR_REG_RDI (arch->current_regs) = method_argument1;
 	INFERIOR_REG_RSI (arch->current_regs) = method_argument2;
@@ -918,10 +956,11 @@ server_ptrace_call_method_1 (ServerHandle *handle, guint64 method_address,
 	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
-	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
+	result = server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_RAX (arch->current_regs) = -1;
 	INFERIOR_REG_RIP (arch->current_regs) = method_address;
 	INFERIOR_REG_RDI (arch->current_regs) = method_argument;
 	INFERIOR_REG_RSI (arch->current_regs) = data_argument;
@@ -987,11 +1026,12 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 		cdata->data_size = data_size;
 	}
 
-	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
+	result = server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
 	g_free (code);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_RAX (arch->current_regs) = -1;
 	INFERIOR_REG_RIP (arch->current_regs) = method_address;
 	INFERIOR_REG_RDI (arch->current_regs) = new_rsp + 8;
 	INFERIOR_REG_RSI (arch->current_regs) = new_rsp + 112;
@@ -1006,6 +1046,69 @@ server_ptrace_call_method_2 (ServerHandle *handle, guint64 method_address,
 	return server_ptrace_continue (handle);
 }
 
+static ServerCommandError
+server_ptrace_call_method_3 (ServerHandle *handle, guint64 method_address,
+			     guint64 method_argument, guint64 address_argument,
+			     guint32 blob_size, gconstpointer blob_data,
+			     guint64 callback_argument)
+{
+	ServerCommandError result = COMMAND_ERROR_NONE;
+	ArchInfo *arch = handle->arch;
+	CallbackData *cdata;
+	guint64 new_rsp;
+	int i;
+
+	static guint8 static_code[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+					0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+					0xcc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
+	int static_size = sizeof (static_code);
+	int size = static_size + blob_size;
+	guint8 *code = g_malloc0 (size);
+	guint64 *ptr = (guint64 *) (code + static_size + blob_size);
+	guint64 effective_address;
+	guint64 blob_start;
+	memcpy (code, static_code, static_size);
+	memcpy (code + static_size, blob_data, blob_size);
+
+	new_rsp = INFERIOR_REG_RSP (arch->current_regs) - AMD64_RED_ZONE_SIZE - size - 16;
+	new_rsp &= 0xfffffffffffffff0L;
+
+	blob_start = new_rsp + static_size;
+
+	*((guint64 *) code) = new_rsp + static_size - 1;
+	*((guint64 *) (code+8)) = callback_argument;
+
+	effective_address = address_argument ? address_argument : blob_start;
+
+	cdata = g_new0 (CallbackData, 1);
+	memcpy (&cdata->saved_regs, &arch->current_regs, sizeof (arch->current_regs));
+	memcpy (&cdata->saved_fpregs, &arch->current_fpregs, sizeof (arch->current_fpregs));
+	cdata->call_address = new_rsp + static_size - 1;
+	cdata->stack_pointer = new_rsp + 8;
+	cdata->callback_argument = callback_argument;
+	cdata->saved_signal = handle->inferior->last_signal;
+	handle->inferior->last_signal = 0;
+
+	result = server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
+	g_free (code);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	INFERIOR_REG_ORIG_RAX (arch->current_regs) = -1;
+	INFERIOR_REG_RIP (arch->current_regs) = method_address;
+	INFERIOR_REG_RDI (arch->current_regs) = method_argument;
+	INFERIOR_REG_RSI (arch->current_regs) = effective_address;
+	INFERIOR_REG_RSP (arch->current_regs) = new_rsp;
+
+	g_ptr_array_add (arch->callback_stack, cdata);
+
+	result = _server_ptrace_set_registers (handle->inferior, &arch->current_regs);
+	if (result != COMMAND_ERROR_NONE)
+		return result;
+
+	return server_ptrace_continue (handle);
+}
+
 static ServerCommandError
 server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 				  guint64 method_argument, guint32 num_params,
@@ -1057,11 +1160,12 @@ server_ptrace_call_method_invoke (ServerHandle *handle, guint64 invoke_method,
 	cdata->saved_signal = handle->inferior->last_signal;
 	handle->inferior->last_signal = 0;
 
-	server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
+	result = server_ptrace_write_memory (handle, (unsigned long) new_rsp, size, code);
 	g_free (code);
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_RAX (arch->current_regs) = -1;
 	INFERIOR_REG_RIP (arch->current_regs) = invoke_method;
 	INFERIOR_REG_RDI (arch->current_regs) = method_argument;
 	INFERIOR_REG_RSI (arch->current_regs) = ptr [0];
@@ -1134,6 +1238,7 @@ server_ptrace_execute_instruction (ServerHandle *handle, const guint8 *instructi
 	if (result != COMMAND_ERROR_NONE)
 		return result;
 
+	INFERIOR_REG_ORIG_RAX (handle->arch->current_regs) = -1;
 	INFERIOR_REG_RIP (handle->arch->current_regs) = code_address;
 
 	result = _server_ptrace_set_registers (handle->inferior, &handle->arch->current_regs);
