@@ -16,6 +16,7 @@ namespace Mono.Debugger
 		RuntimeInvoke,
 		NativeToManaged,
 		ManagedToNative,
+		ManagedToManaged,
 		RemotingInvoke,
 		RemotingInvokeWithCheck,
 		XDomainInvoke,
@@ -27,21 +28,34 @@ namespace Mono.Debugger
 		Synchronized,
 		DynamicMethod,
 		IsInst,
-		CastClass,
+		CanCast,
 		ProxyIsInst,
 		StelemRef,
 		UnBox,
-		Unknown
+		LdFldA,
+		WriteBarrier,
+		Unknown,
+		ComInteropInvoke,
+		ComInterop,
+		Alloc
 	}
 
 	public struct LineEntry : IComparable {
 		public readonly TargetAddress Address;
+		public readonly int File;
 		public readonly int Line;
+		public readonly bool IsHidden;
 
-		public LineEntry (TargetAddress address, int line)
+		public LineEntry (TargetAddress address, int file, int line)
+			: this (address, file, line, false)
+		{ }
+
+		public LineEntry (TargetAddress address, int file, int line, bool is_hidden)
 		{
 			this.Address = address;;
+			this.File = file;
 			this.Line = line;
+			this.IsHidden = is_hidden;
 		}
 
 		public int CompareTo (object obj)
@@ -58,7 +72,7 @@ namespace Mono.Debugger
 
 		public override string ToString ()
 		{
-			return String.Format ("LineEntry ({0}:{1})", Line, Address);
+			return String.Format ("LineEntry ({0}:{1}:{2})", File, Line, Address);
 		}
 	}
 
@@ -170,6 +184,14 @@ namespace Mono.Debugger
 			get;
 		}
 
+		public abstract bool IsWrapper {
+			get;
+		}
+
+		public abstract bool IsCompilerGenerated {
+			get;
+		}
+
 		public abstract bool HasSource {
 			get;
 		}
@@ -262,7 +284,7 @@ namespace Mono.Debugger
                         return true;
                 }
 
-		public abstract TargetClassType GetDeclaringType (Thread target);
+		public abstract TargetStructType GetDeclaringType (Thread target);
 
 		public abstract bool HasThis {
 			get;
@@ -279,6 +301,16 @@ namespace Mono.Debugger
 		internal abstract MethodSource GetTrampoline (TargetMemoryAccess memory,
 							      TargetAddress address);
 
+		internal virtual bool IsIterator {
+			get { return false; }
+		}
+
+		internal virtual Block LookupBlock (TargetMemoryAccess memory,
+						    TargetAddress address)
+		{
+			return null;
+		}
+
 		//
 		// ISourceLookup
 		//
