@@ -451,16 +451,14 @@ namespace Mono.Debugger
 		{
 			check_alive ();
 			servant.Stop ();
-			Wait ();
+			operation_completed_event.WaitOne ();
 			is_running = false;
 			flags |= Flags.AutoRun;
 		}
 
-		public void Wait ()
+		public ThreadCommandResult GetWaitHandle ()
 		{
-			Report.Debug (DebugFlags.Wait, "{0} waiting", this);
-			operation_completed_event.WaitOne ();
-			Report.Debug (DebugFlags.Wait, "{0} done waiting", this);
+			return new ThreadCommandResult (this);
 		}
 
 		// <summary>
@@ -545,7 +543,7 @@ namespace Mono.Debugger
 		}
 
 		public RuntimeInvokeResult RuntimeInvoke (TargetFunctionType function,
-							  TargetClassObject object_argument,
+							  TargetStructObject object_argument,
 							  TargetObject[] param_objects,
 							  bool is_virtual, bool debug)
 		{
@@ -613,6 +611,25 @@ namespace Mono.Debugger
 			return (TargetAddress) result.Result;
 		}
 
+		public TargetAddress CallMethod (TargetAddress method, TargetAddress method_argument,
+						 TargetObject object_argument)
+		{
+			CommandResult result;
+
+			lock (this) {
+				check_alive ();
+				result = servant.CallMethod (
+					method, method_argument, object_argument);
+			}
+
+			result.Wait ();
+
+			if (result.Result == null)
+				throw new TargetException (TargetError.UnknownError);
+
+			return (TargetAddress) result.Result;
+		}
+
 		public void Return (bool run_finally)
 		{
 			CommandResult result;
