@@ -31,7 +31,6 @@ namespace Mono.Debugger.Frontend.CSharp
 		int putback_char;
 		Object val;
 
-		ScriptingContext context;
 		TextReader reader;
 		string ref_name;
 		int current_token;
@@ -46,7 +45,8 @@ namespace Mono.Debugger.Frontend.CSharp
 		// Details about the error encoutered by the tokenizer
 		//
 		string error_details;
-		
+
+		public bool ReadGenericArity { get; set; }
 
 		//
 		// Class initializer
@@ -72,6 +72,7 @@ namespace Mono.Debugger.Frontend.CSharp
 			keywords.Add ("True", Token.TRUE);
 			keywords.Add ("False", Token.FALSE);
 			keywords.Add ("null", Token.NULL);
+			keywords.Add ("@parent", Token.PARENT);
 		}
 
 		public string error {
@@ -80,9 +81,8 @@ namespace Mono.Debugger.Frontend.CSharp
 			}
 		}
 
-		public Tokenizer (ScriptingContext context, TextReader reader, string name)
+		public Tokenizer (TextReader reader, string name)
 		{
-			this.context = context;
 			this.reader = reader;
 			this.ref_name = name;
 		}
@@ -180,8 +180,8 @@ namespace Mono.Debugger.Frontend.CSharp
 					goto default;
 				return v;
 			default:
-				context.Error ("Unrecognized escape sequence in " + (char)d);
-				return d;
+				throw new ScriptingException (
+					"Unrecognized escape sequence in " + (char)d);
 			}
 			GetChar ();
 			return v;
@@ -249,6 +249,26 @@ namespace Mono.Debugger.Frontend.CSharp
 			bool isfloat    = false;
 			bool isdouble   = false;
 			bool isdecimal  = false;
+
+			string digit;
+
+			if (ReadGenericArity) {
+				while (Char.IsDigit((char)PeekChar())) {
+					sb.Append((char)GetChar());
+					++col;
+				}
+
+				digit = sb.ToString ();
+
+				try {
+					val = Int32.Parse(digit, NumberStyles.Number);
+					return Token.INT;
+				} catch (Exception) {
+					error_details = String.Format("Can't parse int {0}", digit);
+					val = 0;
+					return Token.ERROR;
+				}
+			}
 			
 			if (ch == '0' && Char.ToUpper((char)PeekChar()) == 'X') {
 				const string hex = "0123456789ABCDEF";
@@ -334,7 +354,7 @@ namespace Mono.Debugger.Frontend.CSharp
 				}
 			}
 			
-			string digit = sb.ToString();
+			digit = sb.ToString();
 			//string stringValue = String.Concat(prefix.ToString(), digit, suffix.ToString());
 			if (isfloat) {
 				try {
@@ -399,22 +419,6 @@ namespace Mono.Debugger.Frontend.CSharp
 				return Token.UINT;
 			}
 
-			double d = 0;
-			if (!Double.TryParse(digit, ishex ? NumberStyles.HexNumber : NumberStyles.Integer, null, out d)) {
-				error_details = String.Format("Can't parse integral constant {0}", digit);
-				val = 0;
-				return Token.ERROR;
-			}
-			if (d < long.MinValue || d > long.MaxValue) {
-				islong = true;
-				isunsigned = true;	
-			}
-			else if (d < uint.MinValue || d > uint.MaxValue) {
-				islong = true;	
-			}
-			else if (d < int.MinValue || d > int.MaxValue) {
-				isunsigned = true;	
-			}
 			if (islong) {
 				if (isunsigned) {
 					try {
@@ -486,9 +490,9 @@ namespace Mono.Debugger.Frontend.CSharp
 
 			if (keyword == -1 || quoted){
 				val = ids;
-				if (ids.Length > 512){
-					context.Error ("Identifier too long (limit is 512 chars)");
-				}
+				if (ids.Length > 512)
+					throw new ScriptingException (
+						"Identifier too long (limit is 512 chars)");
 				return Token.IDENTIFIER;
 			}
 
@@ -514,7 +518,7 @@ namespace Mono.Debugger.Frontend.CSharp
 
 				if (c == '\n'){
 					if (!quoted)
-						context.Error ("Newline in constant");
+						throw new ScriptingException ("Newline in constant");
 					col = 0;
 				} else
 					col++;
@@ -527,8 +531,7 @@ namespace Mono.Debugger.Frontend.CSharp
 				string_builder.Append ((char) c);
 			}
 
-			context.Error ("Unterminated string literal");
-			return Token.EOF;
+			throw new ScriptingException ("Unterminated string literal");
 		}
 
 		private int consume_quoted_identifier ()
@@ -551,8 +554,7 @@ namespace Mono.Debugger.Frontend.CSharp
 				id_builder.Append ((char) c);
 			}
 
-			context.Error ("Unterminated quoted identifier");
-			return Token.EOF;
+			throw new ScriptingException ("Unterminated quoted identifier");
 		}
 
 		public int xtoken ()
@@ -580,12 +582,26 @@ namespace Mono.Debugger.Frontend.CSharp
 					return ReadDigit ((char)c);
 				}
 
-				if (c == '#')
-					return Token.HASH;
-				else if (c == '@')
+				if (c == '$') {
+					id_builder.Length = 0;
+					
+					while ((c = PeekChar ()) != -1) {
+						if (Char.IsLetterOrDigit ((char)c) || c == '_') {
+							id_builder.Append ((char)GetChar ());
+							col++;
+						} else 
+							break;
+					}
+
+					string ids = id_builder.ToString ();
+					if (ids == "parent")
+						return Token.PARENT;
+					else
+						throw new ScriptingException ("Unknown #-expression");
+				} else if (c == '@')
 					return Token.AT;
-				else if (c == '$')
-					return Token.DOLLAR;
+				else if (c == '#')
+					return Token.HASH;
 				else if (c == '.') {
 					if (Char.IsDigit ((char)PeekChar())) {
 						putback(c);
@@ -633,15 +649,20 @@ namespace Mono.Debugger.Frontend.CSharp
 					return Token.PLUS;
 				}
 				else if (c == '-') { // FIXME: negative numbers...
-					if ((c = PeekChar ()) == '=') {
+					c = PeekChar ();
+					if (c == '=') {
 						GetChar ();
 						return Token.MINUSASSIGN;
 					}
-					if ((c = PeekChar ()) == '>') {
+					if (c == '>') {
 						GetChar ();
 						return Token.ARROW;
 					}
 
+					if (Char.IsDigit ((char) c)) {
+						return ReadDigit ('-');
+					}
+
 					return Token.MINUS;
 				}
 				else if (c == '/') {
@@ -728,6 +749,8 @@ namespace Mono.Debugger.Frontend.CSharp
 					return Token.COLON;
 				else if (c == '?')
 					return Token.QUESTION;
+				else if (c == '`')
+					return Token.BACKTICK;
 
 				if (c == ' ' || c == '\t' || c == '\f' || c == '\v' || c == '\r' || c == '\n'){
 					if (current_token == Token.HASH) {
