@@ -6,6 +6,7 @@ using System.Collections;
 using System.Globalization;
 using Mono.Debugger;
 using Mono.Debugger.Languages;
+using Mono.Debugger.Languages.Mono;
 
 namespace Mono.Debugger.Frontend
 {
@@ -27,9 +28,10 @@ namespace Mono.Debugger.Frontend
 			return parser.Parse (expression);
 		}
 
-		protected static SourceLocation FindFile (Thread target, string filename, int line)
+		protected static SourceLocation FindFile (ScriptingContext context, string filename,
+							  int line)
 		{
-			SourceFile file = target.Process.FindFile (filename);
+			SourceFile file = context.Interpreter.Session.FindFile (filename);
 			if (file == null)
 				throw new ScriptingException ("Cannot find source file `{0}'.",
 							      filename);
@@ -43,13 +45,9 @@ namespace Mono.Debugger.Frontend
 			return new SourceLocation (source, file, line);
 		}
 
-		protected SourceLocation DoParseExpression (Thread target, StackFrame frame,
+		protected SourceLocation DoParseExpression (ScriptingContext context,
 							    LocationType type, string arg)
 		{
-			ScriptingContext context = new ScriptingContext (Interpreter);
-			context.CurrentThread = frame.Thread;
-			context.CurrentFrame = frame;
-
 			Expression expr = Interpreter.ExpressionParser.Parse (arg);
 			if (expr == null)
 				throw new ScriptingException ("Cannot resolve expression `{0}'.", arg);
@@ -62,8 +60,8 @@ namespace Mono.Debugger.Frontend
 				return context.FindMethod (arg);
 		}
 
-		public static bool ParseLocation (Thread target, StackFrame frame,
-						  string arg, out SourceLocation location)
+		public static bool ParseLocation (ScriptingContext context, string arg,
+						  out SourceLocation location)
 		{
 			int line;
 			int pos = arg.IndexOf (':');
@@ -75,7 +73,7 @@ namespace Mono.Debugger.Frontend
 					throw new ScriptingException ("Expected filename:line");
 				}
 
-				location = FindFile (frame.Thread, filename, line);
+				location = FindFile (context, filename, line);
 				return true;
 			}
 
@@ -86,33 +84,38 @@ namespace Mono.Debugger.Frontend
 				return false;
 			}
 
+			StackFrame frame = context.CurrentFrame;
 			if ((frame == null) || (frame.SourceLocation == null) ||
 			    (frame.SourceLocation.FileName == null))
 				throw new ScriptingException (
 					"Current stack frame doesn't have source code");
 
-			location = FindFile (frame.Thread, frame.SourceLocation.FileName, line);
+			location = FindFile (context, frame.SourceLocation.FileName, line);
 			return true;
 		}
 
-		protected SourceLocation DoParse (Thread target, StackFrame frame,
-						  LocationType type, string arg)
+		protected SourceLocation DoParse (ScriptingContext context, LocationType type,
+						  string arg)
 		{
 			if (type != LocationType.Default)
-				return DoParseExpression (target, frame, type, arg);
+				return DoParseExpression (context, type, arg);
 
 			SourceLocation location;
-			if (ParseLocation (target, frame, arg, out location))
+			if (ParseLocation (context, arg, out location))
 				return location;
 
-			return DoParseExpression (target, frame, type, arg);
+			return DoParseExpression (context, type, arg);
 		}
 
 		public SourceLocation Parse (Thread target, StackFrame frame,
 					     LocationType type, string arg)
 		{
+			ScriptingContext context = new ScriptingContext (Interpreter);
+			context.CurrentThread = target;
+			context.CurrentFrame = frame;
+
 			try {
-				return DoParse (target, frame, type, arg);
+				return DoParse (context, type, arg);
 			} catch (ScriptingException ex) {
 				throw new TargetException (TargetError.LocationInvalid, ex.Message);
 			}
@@ -247,7 +250,7 @@ namespace Mono.Debugger.Frontend
 			if (obj == null)
 				return null;
 
-			TargetClassObject cobj = obj as TargetClassObject;
+			TargetStructObject cobj = obj as TargetStructObject;
 			if (cobj != null) {
 				TargetObject current;
 				try {
@@ -255,7 +258,13 @@ namespace Mono.Debugger.Frontend
 				} catch {
 					current = null;
 				}
-				if (current != null)
+				if (current == null)
+					return obj;
+
+				TargetClassObject scurrent = current as TargetClassObject;
+				if ((scurrent != null) && scurrent.Type.IsCompilerGenerated)
+					return obj;
+				else
 					return current;
 			}
 
@@ -477,7 +486,9 @@ namespace Mono.Debugger.Frontend
 			StackFrame frame = context.CurrentFrame;
 			if ((frame.Language == null) ||
 			    !frame.Language.CanCreateInstance (val.GetType ()))
-				throw new ScriptingException ("Cannot instantiate value '{0}' in the current frame's language", Name);
+				throw new ScriptingException (
+					"Cannot instantiate value '{0}' in the current frame's " +
+					"language", Name);
 
 			return frame.Language.CreateInstance (frame.Thread, val);
 		}
@@ -592,7 +603,12 @@ namespace Mono.Debugger.Frontend
 
 		protected override object DoEvaluate (ScriptingContext context)
 		{
-			return null;
+			return IntPtr.Zero;
+		}
+
+		protected override TargetType DoEvaluateType (ScriptingContext context)
+		{
+			return context.CurrentLanguage.VoidType;
 		}
 
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
@@ -895,14 +911,14 @@ namespace Mono.Debugger.Frontend
 			if (method == null)
 				return null;
 
-			TargetClassType decl_type = method.GetDeclaringType (context.CurrentThread);
+			TargetStructType decl_type = method.GetDeclaringType (context.CurrentThread);
 			if (decl_type == null)
 				return null;
 
-			TargetClassObject instance = null;
+			TargetStructObject instance = null;
 			if (method.HasThis) {
 				TargetVariable this_var = method.GetThis (context.CurrentThread);
-				instance = (TargetClassObject) this_var.GetObject (frame);
+				instance = (TargetStructObject) this_var.GetObject (frame);
 			}
 
 			member = StructAccessExpression.FindMember (
@@ -946,6 +962,10 @@ namespace Mono.Debugger.Frontend
 				Expression expr = Lookup (context, frame);
 				if (expr != null)
 					return expr;
+
+				TargetAddress address = context.CurrentProcess.LookupSymbol (name);
+				if (!address.IsNull)
+					return new NumberExpression (address.Address);
 			}
 
 			SourceLocation location = context.FindMethod (name);
@@ -1088,7 +1108,8 @@ namespace Mono.Debugger.Frontend
 		}
 
 		public MemberExpression ResolveMemberAccess (ScriptingContext context,
-							     bool allow_instance)
+							     bool allow_instance,
+							     bool for_invocation)
 		{
 			MemberExpression member;
 
@@ -1114,7 +1135,7 @@ namespace Mono.Debugger.Frontend
 			}
 
 			if (lexpr != null) {
-				TargetClassObject sobj = Convert.ToClassObject (
+				TargetStructObject sobj = Convert.ToStructObject (
 					target, lexpr.EvaluateObject (context));
 				if (sobj == null)
 					throw new ScriptingException (
@@ -1122,6 +1143,14 @@ namespace Mono.Debugger.Frontend
 
 				member = StructAccessExpression.FindMember (
 					target, sobj.Type, sobj, name, true, true);
+
+				if ((member == null) && !sobj.Type.IsByRef && for_invocation) {
+					TargetObjectObject boxed = sobj.Type.Language.CreateBoxedObject (target, sobj);
+					TargetClassObject sboxed = boxed.GetClassObject (target);
+					member = StructAccessExpression.FindMember (
+						target, sobj.Type.Language.ObjectType, sboxed, name, false, true);
+				}
+
 				if (member == null)
 					throw new ScriptingException (
 						"Type `{0}' has no member `{1}'",
@@ -1162,7 +1191,7 @@ namespace Mono.Debugger.Frontend
 
 		protected override Expression DoResolve (ScriptingContext context)
 		{
-			return ResolveMemberAccess (context, false);
+			return ResolveMemberAccess (context, false, false);
 		}
 
 		protected override MethodExpression DoResolveMethod (ScriptingContext context,
@@ -1181,7 +1210,7 @@ namespace Mono.Debugger.Frontend
 				member = StructAccessExpression.FindMember (
 					context.CurrentFrame.Thread, ctype, null, ".ctor", false, true);
 			} else {
-				member = ResolveMemberAccess (context, true);
+				member = ResolveMemberAccess (context, true, true);
 			}
 
 			if (member != null)
@@ -1213,7 +1242,7 @@ namespace Mono.Debugger.Frontend
 
 	public abstract class MemberExpression : Expression
 	{
-		public abstract TargetClassObject InstanceObject {
+		public abstract TargetStructObject InstanceObject {
 			get;
 		}
 
@@ -1255,13 +1284,13 @@ namespace Mono.Debugger.Frontend
 
 	public class MethodGroupExpression : MethodExpression
 	{
-		protected readonly TargetClassType stype;
-		protected readonly TargetClassObject instance;
+		protected readonly TargetStructType stype;
+		protected readonly TargetStructObject instance;
 		protected readonly string name;
 		protected readonly TargetFunctionType[] methods;
 		protected readonly bool is_instance, is_static;
 
-		public MethodGroupExpression (TargetClassType stype, TargetClassObject instance,
+		public MethodGroupExpression (TargetStructType stype, TargetStructObject instance,
 					      string name, TargetFunctionType[] methods,
 					      bool is_instance, bool is_static)
 		{
@@ -1278,7 +1307,7 @@ namespace Mono.Debugger.Frontend
 			get { return stype.Name + "." + name; }
 		}
 
-		public override TargetClassObject InstanceObject {
+		public override TargetStructObject InstanceObject {
 			get { return instance; }
 		}
 
@@ -1384,8 +1413,10 @@ namespace Mono.Debugger.Frontend
 		public static bool IsApplicable (ScriptingContext context, TargetFunctionType method,
 						 TargetType[] types, out string error)
 		{
+			TargetMethodSignature sig = method.GetSignature (context.CurrentThread);
+
 			for (int i = 0; i < types.Length; i++) {
-				TargetType param_type = method.ParameterTypes [i];
+				TargetType param_type = sig.ParameterTypes [i];
 
 				if (param_type == types [i])
 					continue;
@@ -1574,7 +1605,6 @@ namespace Mono.Debugger.Frontend
 	public class RegisterExpression : PointerExpression
 	{
 		string name;
-		int register = -1;
 
 		public RegisterExpression (string register)
 		{
@@ -1622,14 +1652,22 @@ namespace Mono.Debugger.Frontend
 
 		protected override bool DoAssign (ScriptingContext context, TargetObject tobj)
 		{
-			TargetFundamentalObject fobj = tobj as TargetFundamentalObject;
-			if (fobj == null)
-				throw new ScriptingException (
-					"Cannot store non-fundamental object `{0}' in " +
-					" a registers", tobj);
+			long value;
 
-			object obj = fobj.GetObject (context.CurrentThread);
-			long value = System.Convert.ToInt64 (obj);
+			TargetPointerObject pobj = tobj as TargetPointerObject;
+			if (pobj != null) {
+				TargetAddress addr = pobj.GetAddress (context.CurrentThread);
+				value = addr.Address;
+			} else {
+				TargetFundamentalObject fobj = tobj as TargetFundamentalObject;
+				if (fobj == null)
+					throw new ScriptingException (
+						"Cannot store non-fundamental object `{0}' in " +
+						"a register", tobj.Type.Name);
+
+				object obj = fobj.GetObject (context.CurrentThread);
+				value = System.Convert.ToInt64 (obj);
+			}
 
 			Register register = context.CurrentFrame.Registers [name];
 			if (register == null)
@@ -1642,13 +1680,13 @@ namespace Mono.Debugger.Frontend
 
 	public class StructAccessExpression : MemberExpression
 	{
-		public readonly TargetClassType Type;
+		public readonly TargetStructType Type;
 		public readonly TargetMemberInfo Member;
-		protected readonly TargetClassObject instance;
+		protected readonly TargetStructObject instance;
 		TargetClass class_info;
 
-		protected StructAccessExpression (TargetClassType type,
-						  TargetClassObject instance,
+		protected StructAccessExpression (TargetStructType type,
+						  TargetStructObject instance,
 						  TargetMemberInfo member)
 		{
 			this.Type = type;
@@ -1661,7 +1699,7 @@ namespace Mono.Debugger.Frontend
 			get { return Member.Name; }
 		}
 
-		public override TargetClassObject InstanceObject {
+		public override TargetStructObject InstanceObject {
 			get { return instance; }
 		}
 
@@ -1777,64 +1815,100 @@ namespace Mono.Debugger.Frontend
 		}
 
 		public static MemberExpression FindMember (Thread target, TargetStructType stype,
-							   TargetClassObject instance, string name,
+							   TargetStructObject instance, string name,
 							   bool search_static, bool search_instance)
 		{
 		again:
+			TargetClass klass = stype.GetClass (target);
+			if (klass != null) {
+				TargetMemberInfo member = klass.FindMember (
+					target, name, search_static, search_instance);
+				if (member != null)
+					return new StructAccessExpression (stype, instance, member);
+
+				ArrayList methods = new ArrayList ();
+				bool is_instance = false;
+				bool is_static = false;
+
+				TargetMethodInfo[] klass_methods = klass.GetMethods (target);
+				if (klass_methods != null) {
+					foreach (TargetMethodInfo method in klass_methods) {
+						if (method.IsStatic && !search_static)
+							continue;
+						if (!method.IsStatic && !search_instance)
+							continue;
+						if (method.Name != name)
+							continue;
+
+						methods.Add (method.Type);
+						if (method.IsStatic)
+							is_static = true;
+						else
+							is_instance = true;
+					}
+				}
+
+				if (methods.Count > 0) {
+					TargetFunctionType[] funcs = new TargetFunctionType [methods.Count];
+					methods.CopyTo (funcs, 0);
+					return new MethodGroupExpression (
+						stype, instance, name, funcs, is_instance, is_static);
+				}
+			}
+
 			TargetClassType ctype = stype as TargetClassType;
-			if (ctype == null)
-				return null;
+			if (ctype != null) {
+				TargetMemberInfo member = ctype.FindMember (
+					name, search_static, search_instance);
 
-			TargetMemberInfo member = ctype.FindMember (
-				name, search_static, search_instance);
+				if (member != null)
+					return new StructAccessExpression (ctype, instance, member);
 
-			if (member != null)
-				return new StructAccessExpression (ctype, instance, member);
+				ArrayList methods = new ArrayList ();
+				bool is_instance = false;
+				bool is_static = false;
 
-			ArrayList methods = new ArrayList ();
-			bool is_instance = false;
-			bool is_static = false;
-
-			if (name == ".ctor") {
-				foreach (TargetMethodInfo method in ctype.Constructors) {
-					if (method.IsStatic)
-						continue;
-					methods.Add (method.Type);
-					is_instance = true;
-				}
-			} else if (name == ".cctor") {
-				foreach (TargetMethodInfo method in ctype.Constructors) {
-					if (!method.IsStatic)
-						continue;
-					methods.Add (method.Type);
-					is_static = true;
-				}
-			} else {
-				foreach (TargetMethodInfo method in ctype.Methods) {
-					if (method.IsStatic && !search_static)
-						continue;
-					if (!method.IsStatic && !search_instance)
-						continue;
-					if (method.Name != name)
-						continue;
-
-					methods.Add (method.Type);
-					if (method.IsStatic)
-						is_static = true;
-					else
+				if (name == ".ctor") {
+					foreach (TargetMethodInfo method in ctype.Constructors) {
+						if (method.IsStatic)
+							continue;
+						methods.Add (method.Type);
 						is_instance = true;
+					}
+				} else if (name == ".cctor") {
+					foreach (TargetMethodInfo method in ctype.Constructors) {
+						if (!method.IsStatic)
+							continue;
+						methods.Add (method.Type);
+						is_static = true;
+					}
+				} else {
+					foreach (TargetMethodInfo method in ctype.Methods) {
+						if (method.IsStatic && !search_static)
+							continue;
+						if (!method.IsStatic && !search_instance)
+							continue;
+						if (method.Name != name)
+							continue;
+
+						methods.Add (method.Type);
+						if (method.IsStatic)
+							is_static = true;
+						else
+							is_instance = true;
+					}
+				}
+
+				if (methods.Count > 0) {
+					TargetFunctionType[] funcs = new TargetFunctionType [methods.Count];
+					methods.CopyTo (funcs, 0);
+					return new MethodGroupExpression (
+						ctype, instance, name, funcs, is_instance, is_static);
 				}
 			}
 
-			if (methods.Count > 0) {
-				TargetFunctionType[] funcs = new TargetFunctionType [methods.Count];
-				methods.CopyTo (funcs, 0);
-				return new MethodGroupExpression (
-					ctype, instance, name, funcs, is_instance, is_static);
-			}
-
-			if (ctype.HasParent) {
-				stype = ctype.GetParentType (target);
+			if (stype.HasParent) {
+				stype = stype.GetParentType (target);
 				if (instance != null) {
 					instance = instance.GetParentObject (target);
 					if (instance == null)
@@ -1931,6 +2005,23 @@ namespace Mono.Debugger.Frontend
 			class_info.SetField (target, InstanceObject, field, obj);
 		}
 
+		protected void SetProperty (ScriptingContext context, TargetPropertyInfo prop,
+					    TargetObject obj)
+		{
+			ResolveClass (context.CurrentThread);
+			if (prop.Setter == null)
+				throw new ScriptingException ("Property `{0}' has no setter.", Name);
+
+			RuntimeInvokeResult result = context.Interpreter.RuntimeInvoke (
+				context.CurrentThread, prop.Setter, InstanceObject,
+				new TargetObject [] { obj }, true, false);
+
+			if (result.ExceptionMessage != null)
+				throw new ScriptingException (
+					"Invocation of `{0}' raised an exception: {1}",
+					Name, result.ExceptionMessage);
+		}
+
 		protected override bool DoAssign (ScriptingContext context, TargetObject obj)
 		{
 			if (Member is TargetFieldInfo) {
@@ -1941,10 +2032,15 @@ namespace Mono.Debugger.Frontend
 						obj.TypeName, Name, Member.Type.Name);
 
 				SetField (context.CurrentThread, (TargetFieldInfo) Member, obj);
-			}
-			else if (Member is TargetPropertyInfo) 
-			  	throw new ScriptingException ("Can't set properties directly.");
-			else if (Member is TargetEventInfo)
+			} else if (Member is TargetPropertyInfo) {
+				if (Member.Type != obj.Type)
+					throw new ScriptingException (
+						"Type mismatch: cannot assign expression of type " +
+						"`{0}' to property `{1}', which is of type `{2}'.",
+						obj.TypeName, Name, Member.Type.Name);
+
+				SetProperty (context, (TargetPropertyInfo) Member, obj);
+			} else if (Member is TargetEventInfo)
 				throw new ScriptingException ("Can't set events directly.");
 			else if (Member is TargetMethodInfo)
 				throw new ScriptingException ("Can't set methods directly.");
@@ -1987,7 +2083,7 @@ namespace Mono.Debugger.Frontend
 			TargetType type = expr.EvaluateType (context);
 
 			TargetPointerType ptype = type as TargetPointerType;
-			if (ptype != null)
+			if ((ptype != null) && ptype.CanDereference)
 				return ptype.StaticType;
 
 			throw new ScriptingException (
@@ -1999,22 +2095,19 @@ namespace Mono.Debugger.Frontend
 			TargetObject obj = expr.EvaluateObject (context);
 
 			TargetPointerObject pobj = obj as TargetPointerObject;
-			if (pobj != null) {
+			if ((pobj != null) && pobj.Type.CanDereference) {
 				TargetObject result;
 				try {
 					result = pobj.GetDereferencedObject (context.CurrentThread);
 				} catch {
-					result = null;
+					throw new ScriptingException ("Cannot dereference `{0}'.",
+								      expr.Name);
 				}
 
 				if (result != null)
 					return result;
 			}
 
-			TargetClassObject cobj = obj as TargetClassObject;
-			if (current_ok && (cobj != null))
-				return cobj;
-
 			PointerExpression pexpr = expr as PointerExpression;
 			if (pexpr != null) {
 				TargetAddress address = pexpr.EvaluateAddress (context);
@@ -2029,6 +2122,10 @@ namespace Mono.Debugger.Frontend
 				}
 			}
 
+			TargetClassObject cobj = obj as TargetClassObject;
+			if (current_ok && (cobj != null))
+				return cobj;
+
 			throw new ScriptingException (
 				"Expression `{0}' is not a pointer.", expr.Name);
 		}
@@ -2040,8 +2137,12 @@ namespace Mono.Debugger.Frontend
 				obj = (long) (int) obj;
 			if (obj is long)
 				return new TargetAddress (context.AddressDomain, (long) obj);
+			else if (obj is PointerDereferenceExpression)
+				obj = ((Expression) obj).EvaluateObject (context);
 			else if (obj is PointerExpression)
 				return ((PointerExpression) obj).EvaluateAddress (context);
+			else if (obj is Expression)
+				obj = ((Expression) obj).EvaluateObject (context);
 
 			TargetPointerObject pobj = obj as TargetPointerObject;
 			if (pobj == null)
@@ -2164,6 +2265,8 @@ namespace Mono.Debugger.Frontend
 
 				if (idx is int)
 					return (int) idx;
+				else if (idx is long)
+					return (int) (long) idx;
 
 				TargetFundamentalObject obj = (TargetFundamentalObject) idx;
 				return (int) obj.GetObject (target);
@@ -2353,10 +2456,10 @@ namespace Mono.Debugger.Frontend
 			return this;
 		}
 
-		static TargetClassObject TryParentCast (ScriptingContext context,
-							TargetClassObject source,
-							TargetStructType source_type,
-							TargetStructType target_type)
+		static TargetStructObject TryParentCast (ScriptingContext context,
+							 TargetStructObject source,
+							 TargetStructType source_type,
+							 TargetStructType target_type)
 		{
 			if (source_type == target_type)
 				return source;
@@ -2369,14 +2472,14 @@ namespace Mono.Debugger.Frontend
 			if (source == null)
 				return null;
 
-			return source.GetParentObject (context.CurrentThread);
+			return source.GetParentObject (context.CurrentThread) as TargetClassObject;
 		}
 
-		static TargetClassObject TryCurrentCast (ScriptingContext context,
-							 TargetClassObject source,
-							 TargetClassType target_type)
+		static TargetStructObject TryCurrentCast (ScriptingContext context,
+							  TargetClassObject source,
+							  TargetClassType target_type)
 		{
-			TargetClassObject current = source.GetCurrentObject (context.CurrentThread);
+			TargetStructObject current = source.GetCurrentObject (context.CurrentThread);
 			if (current == null)
 				return null;
 
@@ -2393,8 +2496,7 @@ namespace Mono.Debugger.Frontend
 			if (sobj == null)
 				return null;
 
-			TargetClassObject result;
-			result = TryParentCast (context, sobj, sobj.Type, target_type);
+			TargetStructObject result = TryParentCast (context, sobj, sobj.Type, target_type);
 			if (result != null)
 				return result;
 
@@ -2432,6 +2534,30 @@ namespace Mono.Debugger.Frontend
 		{
 			TargetType target_type = target.EvaluateType (context);
 
+			TargetObject obj = DoCast (context, expr, target_type);
+			if (obj == null)
+				throw new ScriptingException (
+					"Cannot cast from `{0}' to `{1}'.", expr.Name, target.Name);
+
+			return obj;
+		}
+
+		TargetObject DoCast (ScriptingContext context, Expression expr,
+				     TargetType target_type)
+		{
+			TargetObject source = expr.EvaluateObject (context);
+			if (source == null)
+				return null;
+
+			if (target_type is TargetObjectType) {
+				if (((source is TargetClassObject) && !source.Type.IsByRef) ||
+				    (source is TargetFundamentalObject))
+					return target_type.Language.CreateBoxedObject (context.CurrentThread, source);
+
+				throw new ScriptingException (
+					"Cannot box object `{0}': not a value-type", expr.Name);
+			}
+
 			if (target_type is TargetPointerType) {
 				TargetAddress address;
 
@@ -2439,13 +2565,10 @@ namespace Mono.Debugger.Frontend
 				if (pexpr != null)
 					address = pexpr.EvaluateAddress (context);
 				else {
-					TargetType etype = expr.EvaluateType (context);
-					TargetPointerType ptype = etype as TargetPointerType;
-
+					TargetPointerType ptype = expr.EvaluateType (context)
+						as TargetPointerType;
 					if ((ptype == null) || ptype.IsTypesafe)
-						throw new ScriptingException (
-							"Cannot cast from {0} to {1}.",
-							etype.Name, target.Name);
+						return null;
 
 					pexpr = new AddressOfExpression (expr);
 					pexpr.Resolve (context);
@@ -2456,21 +2579,24 @@ namespace Mono.Debugger.Frontend
 				return ((TargetPointerType) target_type).GetObject (address);
 			}
 
-			TargetClassType type = Convert.ToClassType (target_type);
-			TargetClassObject source = Convert.ToClassObject (
-				context.CurrentThread, expr.EvaluateObject (context));
+			if (target_type is TargetFundamentalType) {
+				TargetFundamentalObject fobj = expr.EvaluateObject (context)
+					as TargetFundamentalObject;
+				if (fobj == null)
+					return null;
 
-			if (source == null)
+				TargetFundamentalType ftype = target_type as TargetFundamentalType;
+				return Convert.ExplicitFundamentalConversion (context, fobj, ftype);
+			}
+
+			TargetClassType ctype = Convert.ToClassType (target_type);
+			TargetClassObject source_cobj = Convert.ToClassObject (context.CurrentThread, source);
+
+			if (source_cobj == null)
 				throw new ScriptingException (
 					"Variable {0} is not a class type.", expr.Name);
 
-			TargetObject obj = TryCast (context, source, type);
-			if (obj == null)
-				throw new ScriptingException (
-					"Cannot cast from {0} to {1}.", source.Type.Name,
-					type.Name);
-
-			return obj;
+			return TryCast (context, source_cobj, ctype);
 		}
 
 		protected override TargetType DoEvaluateType (ScriptingContext context)
@@ -2634,6 +2760,28 @@ namespace Mono.Debugger.Frontend
 			return type.Language.CreateInstance (context.CurrentThread, new_value);
 		}
 
+		public static TargetObject ExplicitFundamentalConversion (ScriptingContext context,
+									  TargetFundamentalObject obj,
+									  TargetFundamentalType type)
+		{
+			TargetObject retval = ImplicitFundamentalConversion (context, obj, type);
+			if (retval != null)
+				return retval;
+
+			FundamentalKind tkind = type.FundamentalKind;
+
+			try {
+				object value = obj.GetObject (context.CurrentThread);
+				object new_value = ImplicitFundamentalConversion (value, tkind);
+				if (new_value == null)
+					return null;
+
+				return type.Language.CreateInstance (context.CurrentThread, new_value);
+			} catch {
+				return null;
+			}
+		}
+
 		static bool ImplicitReferenceConversionExists (ScriptingContext context,
 							       TargetStructType source,
 							       TargetStructType target)
@@ -2749,6 +2897,23 @@ namespace Mono.Debugger.Frontend
 
 			return null;
 		}
+
+		public static TargetStructObject ToStructObject (Thread target, TargetObject obj)
+		{
+			TargetStructObject sobj = obj as TargetStructObject;
+			if (sobj != null)
+				return sobj;
+
+			TargetObjectObject oobj = obj as TargetObjectObject;
+			if (oobj != null)
+				return oobj.GetClassObject (target);
+
+			TargetArrayObject aobj = obj as TargetArrayObject;
+			if ((aobj != null) && aobj.HasClassObject)
+				return aobj.GetClassObject (target);
+
+			return null;
+		}
 	}
 
 	public class ConditionalExpression : Expression
@@ -2878,7 +3043,7 @@ namespace Mono.Debugger.Frontend
 			return mg;
 		}
 
-		public override TargetClassObject InstanceObject {
+		public override TargetStructObject InstanceObject {
 			get { return method_expr.InstanceObject; }
 		}
 
@@ -2971,13 +3136,15 @@ namespace Mono.Debugger.Frontend
 			for (int i = 0; i < arguments.Length; i++)
 				args [i] = arguments [i].EvaluateObject (context);
 
+			TargetMethodSignature sig = method.GetSignature (context.CurrentThread);
+
 			TargetObject[] objs = new TargetObject [args.Length];
 			for (int i = 0; i < args.Length; i++) {
 				objs [i] = Convert.ImplicitConversionRequired (
-					context, args [i], method.ParameterTypes [i]);
+					context, args [i], sig.ParameterTypes [i]);
 			}
 
-			TargetClassObject instance = method_expr.InstanceObject;
+			TargetStructObject instance = method_expr.InstanceObject;
 
 			if (!method.IsStatic && !method.IsConstructor && (instance == null))
 				throw new ScriptingException (
@@ -3122,4 +3289,56 @@ namespace Mono.Debugger.Frontend
 			return obj;
 		}
 	}
+
+	public class ParentExpression : Expression
+	{
+		Expression expr;
+		int level;
+		string name;
+
+		public ParentExpression (Expression expr, int level)
+		{
+			this.expr = expr;
+			this.level = level;
+
+			if (level > 0)
+				name = String.Format ("$parent+{0} ({1})", level, expr.Name);
+			else
+				name = String.Format ("$parent ({0})", expr.Name);
+		}
+
+		public override string Name {
+			get { return name; }
+		}
+
+		protected override Expression DoResolve (ScriptingContext context)
+		{
+			expr = expr.Resolve (context);
+			if (expr == null)
+				return null;
+
+			resolved = true;
+			return this;
+		}
+
+		protected override TargetObject DoEvaluateObject (ScriptingContext context)
+		{
+			TargetVariable var = expr.EvaluateVariable (context);
+			if (var == null)
+				return null;
+
+			TargetObject obj = var.GetObject (context.CurrentFrame);
+			if (obj == null)
+				return null;
+
+			TargetStructObject sobj = (TargetStructObject) obj;
+			for (int i = 0; i <= level; i++) {
+				sobj = sobj.GetParentObject (context.CurrentThread);
+				if (sobj == null)
+					return null;
+			}
+
+			return sobj;
+		}
+	}
 }
