@@ -1,10 +1,10 @@
 using System;
-using Math = System.Math;
 using System.Text;
 using System.IO;
 using System.Threading;
 using System.Reflection;
 using System.Collections;
+using System.Collections.Generic;
 using System.Globalization;
 using System.Runtime.Serialization;
 using System.Runtime.InteropServices;
@@ -39,6 +39,7 @@ namespace Mono.Debugger.Frontend
 
 		ExpressionParser parser;
 		ManualResetEvent interrupt_event;
+		ManualResetEvent process_event;
 		Thread current_thread;
 
 		internal static readonly string DirectorySeparatorStr;
@@ -62,6 +63,7 @@ namespace Mono.Debugger.Frontend
 			source_factory = new SourceFileFactory ();
 
 			interrupt_event = new ManualResetEvent (false);
+			process_event = new ManualResetEvent (false);
 
 			styles = new Hashtable ();
 			styles.Add ("cli", new StyleCLI (this));
@@ -70,7 +72,6 @@ namespace Mono.Debugger.Frontend
 		}
 
 		AppDomain debugger_domain;
-		static int domain_age = 0;
 
 		public void Exit ()
 		{
@@ -295,8 +296,7 @@ namespace Mono.Debugger.Frontend
 				current_process = main_process = debugger.Run (session);
 
 				current_thread = current_process.MainThread;
-				Wait (current_thread);
-				CheckLastEvent (current_thread);
+				current_thread = WaitAll (current_thread);
 
 				return current_process;
 			} catch (TargetException) {
@@ -322,8 +322,7 @@ namespace Mono.Debugger.Frontend
 
 				current_process = main_process = debugger.Attach (session, pid);
 				current_thread = current_process.MainThread;
-				Wait (current_thread);
-				CheckLastEvent (current_thread);
+				current_thread = WaitAll (current_thread);
 
 				return current_process;
 			} catch (TargetException) {
@@ -380,8 +379,7 @@ namespace Mono.Debugger.Frontend
 				current_process = main_process = debugger.Run (session);
 
 				current_thread = current_process.MainThread;
-				Wait (current_thread);
-				CheckLastEvent (current_thread);
+				current_thread = WaitAll (current_thread);
 
 				return current_process;
 			} catch (TargetException ex) {
@@ -441,19 +439,27 @@ namespace Mono.Debugger.Frontend
 			return true;
 		}
 
+		public Thread WaitAll (Thread thread)
+		{
+			return WaitAll (thread, thread.GetWaitHandle (), false);
+		}
+
 		public Thread WaitAll (Thread thread, CommandResult result, bool wait)
 		{
 			if (result == null)
 				return null;
 
 			ClearInterrupt ();
+			process_event.Reset ();
 
 			Thread stopped;
 			Hashtable seen_threads = new Hashtable ();
 			Hashtable wait_list = new Hashtable ();
 
 			do {
-				foreach (Process process in Processes) {
+			again:
+				Process[] processes = HasTarget ? debugger.Processes : new Process [0];
+				foreach (Process process in processes) {
 					foreach (Thread t in process.GetThreads ()) {
 						if ((t == thread) || seen_threads.Contains (t))
 							continue;
@@ -468,15 +474,16 @@ namespace Mono.Debugger.Frontend
 					}
 				}
 
-				WaitHandle[] handles = new WaitHandle [wait_list.Count + 2];
+				WaitHandle[] handles = new WaitHandle [wait_list.Count + 3];
 				handles [0] = interrupt_event;
-				handles [1] = result.CompletedEvent;
+				handles [1] = process_event;
+				handles [2] = result.CompletedEvent;
 
 				Thread[] threads = new Thread [wait_list.Count];
 				wait_list.Keys.CopyTo (threads, 0);
 
 				for (int i = 0; i < wait_list.Count; i++)
-					handles [i + 2] = threads [i].WaitHandle;
+					handles [i + 3] = threads [i].WaitHandle;
 
 				int ret = WaitHandle.WaitAny (handles);
 
@@ -485,15 +492,18 @@ namespace Mono.Debugger.Frontend
 					result.Abort ();
 					result.CompletedEvent.WaitOne ();
 					break;
+				} else if (ret == 1) {
+					process_event.Reset ();
+					goto again;
 				}
 
 				if (result.Result is Exception)
 					throw (Exception) result.Result;
 
-				if (ret == 1)
+				if (ret == 2)
 					stopped = thread;
 				else
-					stopped = threads [ret - 2];
+					stopped = threads [ret - 3];
 
 				CheckLastEvent (stopped);
 				wait_list.Remove (stopped);
@@ -537,7 +547,7 @@ namespace Mono.Debugger.Frontend
 
 		public RuntimeInvokeResult RuntimeInvoke (Thread thread,
 							  TargetFunctionType function,
-							  TargetClassObject object_argument,
+							  TargetStructObject object_argument,
 							  TargetObject[] param_objects,
 							  bool is_virtual, bool debug)
 		{
@@ -550,11 +560,12 @@ namespace Mono.Debugger.Frontend
 					result.Abort ();
 					return null;
 				}
-			} else {
+			} else if (debug) {
 				Wait (thread);
 
-				if (debug)
-					CheckLastEvent (thread);
+				CheckLastEvent (thread);
+			} else {
+				Wait (result);
 			}
 
 			return result;
@@ -610,23 +621,32 @@ namespace Mono.Debugger.Frontend
 
 		protected virtual void OnThreadCreated (Thread thread)
 		{
-			Print ("Process #{0} created new thread @{1}.",
-			       thread.Process.ID, thread.ID);
+			process_event.Set ();
+			if (DebuggerConfiguration.NotifyUser_ThreadCreation)
+				Print ("Process #{0} created new thread @{1}.",
+				       thread.Process.ID, thread.ID);
 		}
 
 		protected virtual void OnThreadExited (Thread thread)
 		{
-			if (thread != thread.Process.MainThread)
+			if (DebuggerConfiguration.NotifyUser_ThreadCreation)
 				Print ("Thread @{0} exited.", thread.ID);
 			if (thread == current_thread)
 				current_thread = null;
 		}
 
 		protected virtual void OnMainProcessCreated (Process process)
+		{
+			new ProcessEventSink (this, process);
+		}
+
+		protected virtual void OnProcessReachedMain (Process process)
 		{ }
 
 		protected virtual void OnProcessCreated (Process process)
 		{
+			process_event.Set ();
+			new ProcessEventSink (this, process);
 			Print ("Created new process #{0}.", process.ID);
 			if (current_process == null) {
 				current_process = process;
@@ -847,24 +867,24 @@ namespace Mono.Debugger.Frontend
 
 		public Module[] GetModules (int[] indices)
 		{
-			int pos = 0;
-			Module[] retval = new Module [indices.Length];
+			List<Module> retval = new List<Module> ();
 			Module[] modules = CurrentProcess.Modules;
 
 			foreach (int index in indices) {
-				if ((index < 0) || (index > modules.Length))
-					throw new ScriptingException ("No such module {0}.", index);
-
+				bool found = false;
 				for (int i = 0; i < modules.Length; i++) {
-					if (modules [i].ID != index)
-						continue;
-					retval [pos] = modules [i];
+					if (modules [i].ID == index) {
+						retval.Add (modules [i]);
+						found = true;
+						break;
+					}
 				}
 
-				pos ++;
+				if (!found)
+					throw new ScriptingException ("No such module {0}.", index);
 			}
 
-			return retval;
+			return retval.ToArray ();
 		}
 
 		public SourceFile[] GetSources (int[] indices)
@@ -1021,10 +1041,10 @@ namespace Mono.Debugger.Frontend
 				debugger.ThreadCreatedEvent += thread_created;
 				debugger.ThreadExitedEvent += thread_exited;
 				debugger.MainProcessCreatedEvent += main_process_created;
+				debugger.ProcessReachedMainEvent += process_reached_main;
 				debugger.ProcessCreatedEvent += process_created;
 				debugger.ProcessExitedEvent += process_exited;
 				debugger.ProcessExecdEvent += process_execd;
-				debugger.TargetOutputEvent += target_output;
 			}
 
 			public void thread_created (Debugger debugger, Thread thread)
@@ -1044,6 +1064,11 @@ namespace Mono.Debugger.Frontend
 				interpreter.OnMainProcessCreated (process);
 			}
 
+			public void process_reached_main (Debugger debugger, Process process)
+			{
+				interpreter.OnProcessReachedMain (process);
+			}
+
 			public void process_created (Debugger debugger, Process process)
 			{
 				interpreter.OnProcessCreated (process);
@@ -1063,6 +1088,17 @@ namespace Mono.Debugger.Frontend
 			{
 				interpreter.OnTargetExited ();
 			}
+		}
+
+		protected class ProcessEventSink : DebuggerMarshalByRefObject
+		{
+			Interpreter interpreter;
+
+			public ProcessEventSink (Interpreter interpreter, Process process)
+			{
+				this.interpreter = interpreter;				
+				process.TargetOutputEvent += target_output;
+			}
 
 			public void target_output (bool is_stderr, string line)
 			{
