@@ -235,11 +235,18 @@ namespace Mono.Debugger.Frontend
 		{
 			bool is_terminal = GnuReadLine.IsTerminal (0);
 
-			DebuggerConfiguration config = new DebuggerConfiguration ();
-			config.LoadConfiguration ();
-
 			DebuggerOptions options = DebuggerOptions.ParseCommandLine (args);
 
+			DebuggerConfiguration config = new DebuggerConfiguration ();
+#if HAVE_XSP
+			if (options.StartXSP)
+				config.SetupXSP ();
+			else
+				config.LoadConfiguration ();
+#else
+			config.LoadConfiguration ();
+#endif
+
 			Console.WriteLine ("Mono Debugger");
 
 			CommandLineInterpreter interpreter = new CommandLineInterpreter (
