@@ -171,15 +171,20 @@ namespace Mono.Debugger.Frontend
 			interpreter.Print (obj);
 		}
 
-		string MonoObjectToString (TargetClassObject obj)
+		string MonoObjectToString (TargetStructObject obj)
 		{
-			TargetClassObject cobj = obj;
+			TargetStructObject cobj = obj;
 
 		again:
-			TargetClassType ctype = cobj.Type;
+			TargetStructType ctype = cobj.Type;
 			if ((ctype.Name == "System.Object") || (ctype.Name == "System.ValueType"))
 				return null;
-			TargetMethodInfo[] methods = ctype.Methods;
+
+			TargetClass klass = ctype.GetClass (CurrentThread);
+			TargetMethodInfo[] methods = klass.GetMethods (CurrentThread);
+			if (methods == null)
+				return null;
+
 			foreach (TargetMethodInfo minfo in methods) {
 				if (minfo.Name != "ToString")
 					continue;
@@ -202,7 +207,7 @@ namespace Mono.Debugger.Frontend
 				return String.Format ("({0}) {{ \"{1}\" }}", obj.Type.Name, value);
 			}
 
-			cobj = cobj.GetParentObject (CurrentThread);
+			cobj = cobj.GetParentObject (CurrentThread) as TargetStructObject;
 			if (cobj != null)
 				goto again;
 
@@ -212,7 +217,7 @@ namespace Mono.Debugger.Frontend
 		string DoFormatObject (TargetObject obj, DisplayFormat format)
 		{
 			if (format == DisplayFormat.Object) {
-				TargetClassObject cobj = obj as TargetClassObject;
+				TargetStructObject cobj = obj as TargetStructObject;
 				if (cobj != null) {
 					string formatted = MonoObjectToString (cobj);
 					if (formatted != null)
@@ -330,18 +335,36 @@ namespace Mono.Debugger.Frontend
 
 		public SourceLocation FindMethod (string name)
 		{
-			return CurrentProcess.FindMethod (name);
+			foreach (Module module in CurrentProcess.Modules) {
+				MethodSource method = module.FindMethod (name);
+				
+				if (method != null)
+					return new SourceLocation (method);
+			}
+
+			return null;
 		}
 
 		public void ShowSources (Module module)
 		{
-			if (!module.SymbolsLoaded)
+			if (!module.SymbolsLoaded) {
+				Print ("Symbols from module {0} not loaded.", module.Name);
 				return;
+			}
 
 			Print ("Sources for module {0}:", module.Name);
 
-			foreach (SourceFile source in module.Sources)
-				Print ("{0,4}  {1}", source.ID, source.FileName);
+			foreach (SourceFile source in module.Sources) {
+				if (source.IsAutoGenerated &&
+				    interpreter.DebuggerConfiguration.HideAutoGenerated)
+					continue;
+				Print ("{0,4}  {1}{2}", source.ID, source.FileName,
+				       source.IsAutoGenerated ? " (auto-generated)" : "");
+				if (source.CheckModified ()) {
+					Print ("      Source file {0} is more recent than executable.",
+					       source.FileName);
+				}
+			}
 		}
 
 		public void LoadLibrary (Thread thread, string filename)
