@@ -1,6 +1,7 @@
 using System;
 using System.Text;
 using System.Collections;
+using System.Collections.Generic;
 using System.Globalization;
 using Mono.Debugger;
 using Mono.Debugger.Languages;
@@ -109,7 +110,7 @@ namespace Mono.Debugger.Frontend
 			} else
 				buffer = location.SourceBuffer;
 
-			if ((buffer == null) || (buffer.Contents == null))
+			if ((buffer == null) || (buffer.Contents == null) || (location.Row == 0))
 				return false;
 
 			string line = buffer.Contents [location.Row - 1];
@@ -127,8 +128,6 @@ namespace Mono.Debugger.Frontend
 				if ((int) args.Data != 0)
 					interpreter.Print ("{0} exited with exit code {1}.",
 							   thread.Name, (int) args.Data);
-				else
-					interpreter.Print ("{0} exited normally.", thread.Name);
 				break;
 
 			case TargetEventType.TargetSignaled:
@@ -363,99 +362,30 @@ namespace Mono.Debugger.Frontend
 
 			case TargetObjectKind.Class:
 			case TargetObjectKind.Struct: {
-				TargetClassType stype = (TargetClassType) type;
 				StringBuilder sb = new StringBuilder ();
-				TargetClassType ctype = type as TargetClassType;
+				TargetClassType ctype = (TargetClassType) type;
 				if (type.Kind == TargetObjectKind.Struct)
 					sb.Append ("struct ");
 				else
 					sb.Append ("class ");
-				if (ctype != null) {
-					if (ctype.Name != null) {
-						sb.Append (ctype.Name);
-						sb.Append (" ");
-					}
-					if (ctype.HasParent) {
-						TargetStructType parent = ctype.GetParentType (target);
-						sb.Append (": ");
-						sb.Append (parent.Name);
-					}
-				} else {
-					if (stype.Name != null) {
-						sb.Append (stype.Name);
-						sb.Append (" ");
-					}
+				if (ctype.Name != null) {
+					sb.Append (ctype.Name);
+					sb.Append (" ");
 				}
-				sb.Append ("\n" + prefix + "{\n");
-				foreach (TargetFieldInfo field in stype.Fields) {
-					if (field.IsStatic)
-						continue;
-					sb.Append (FormatMember (
-							   prefix, field, false, hash) + ";\n");
-				}
-				foreach (TargetFieldInfo field in stype.Fields) {
-					if (!field.IsStatic)
-						continue;
-					sb.Append (FormatMember (
-							   prefix, field, true, hash) + ";\n");
-				}
-				foreach (TargetPropertyInfo property in stype.Properties) {
-					if (property.IsStatic)
-						continue;
-					sb.Append (FormatProperty (
-							   prefix, property, false, hash));
-				}
-				foreach (TargetPropertyInfo property in stype.Properties) {
-					if (!property.IsStatic)
-						continue;
-					sb.Append (FormatProperty (
-							   prefix, property, true, hash));
-				}
-				foreach (TargetEventInfo ev in stype.Events) {
-					if (ev.IsStatic)
-						continue;
-					sb.Append (FormatEvent (
-							   prefix, ev, false, hash));
-				}
-				foreach (TargetEventInfo ev in stype.Events) {
-					if (!ev.IsStatic)
-						continue;
-					sb.Append (FormatEvent (
-							   prefix, ev, true, hash));
-				}
-				foreach (TargetMethodInfo method in stype.Methods) {
-					if (method.IsStatic)
-						continue;
-					sb.Append (FormatMethod (
-							   prefix, method, false, false, hash));
-				}
-				foreach (TargetMethodInfo method in stype.Methods) {
-					if (!method.IsStatic)
-						continue;
-					sb.Append (FormatMethod (
-							   prefix, method, true, false, hash));
-				}
-				foreach (TargetMethodInfo method in stype.Constructors) {
-					if (method.IsStatic)
-						continue;
-					sb.Append (FormatMethod (
-							   prefix, method, false, true, hash));
-				}
-				foreach (TargetMethodInfo method in stype.Constructors) {
-					if (!method.IsStatic)
-						continue;
-					sb.Append (FormatMethod (
-							   prefix, method, true, true, hash));
+				if (ctype.HasParent) {
+					TargetStructType parent = ctype.GetParentType (target);
+					sb.Append (": ");
+					sb.Append (parent.Name);
 				}
 
-				sb.Append (prefix);
-				sb.Append ("}");
+				sb.Append ("\n" + prefix + "{\n");
+				sb.Append (FormatStruct (prefix, ctype, hash));
+				sb.Append (prefix + "}");
 
 				retval = sb.ToString ();
 				break;
 			}
 
-#if FIXME
 			case TargetObjectKind.Alias: {
 				TargetTypeAlias alias = (TargetTypeAlias) type;
 				string name;
@@ -466,7 +396,30 @@ namespace Mono.Debugger.Frontend
 				retval = String.Format ("typedef {0} = {1}", alias.Name, name);
 				break;
 			}
-#endif
+
+			case TargetObjectKind.GenericInstance: {
+				TargetGenericInstanceType gtype = (TargetGenericInstanceType) type;
+
+				StringBuilder sb = new StringBuilder ();
+				if (gtype.ContainerType.Kind == TargetObjectKind.Struct)
+					sb.Append ("struct ");
+				else
+					sb.Append ("class ");
+
+				sb.Append (String.Format ("{0} = ", gtype.Name));
+
+				TargetStructType parent = gtype.ContainerType.GetParentType (target);
+				sb.Append (String.Format ("{0}", gtype.ContainerType.Name));
+				if (parent != null)
+					sb.Append (String.Format (" : {0}", parent.Name));
+
+				sb.Append ("\n" + prefix + "{\n");
+				sb.Append (FormatStruct (prefix, gtype.ContainerType, hash));
+				sb.Append (prefix + "}");
+
+				retval = sb.ToString ();
+				break;
+			}
 
 			default:
 				retval = type.Name;
@@ -477,6 +430,208 @@ namespace Mono.Debugger.Frontend
 			return retval;
 		}
 
+		protected void FormatAccessibility (StringBuilder sb, string prefix,
+						    TargetMemberAccessibility accessibility)
+		{
+			switch (accessibility) {
+			case TargetMemberAccessibility.Public:
+				sb.Append (prefix + "public:\n");
+				break;
+			case TargetMemberAccessibility.Protected:
+				sb.Append (prefix + "protected:\n");
+				break;
+			case TargetMemberAccessibility.Internal:
+				sb.Append (prefix + "internal:\n");
+				break;
+			default:
+				sb.Append (prefix + "private:\n");
+				break;
+			}
+		}
+
+		protected void FormatFields (TargetClassType type, bool is_static,
+					     TargetMemberAccessibility accessibility,
+					     List<string> members, string prefix, Hashtable hash)
+		{
+			List<TargetFieldInfo> list = new List<TargetFieldInfo> ();
+			foreach (TargetFieldInfo field in type.Fields) {
+				if (field.IsStatic != is_static)
+					continue;
+				if (field.Accessibility != accessibility)
+					continue;
+				list.Add (field);
+			}
+			if (list.Count == 0)
+				return;
+
+			foreach (TargetFieldInfo field in list)
+				members.Add (FormatMember (prefix, field, is_static, hash) + ";\n");
+		}
+
+		protected void FormatFields (TargetClassType type,
+					     TargetMemberAccessibility accessibility,
+					     List<String> members, string prefix, Hashtable hash)
+		{
+			FormatFields (type, false, accessibility, members, prefix, hash);
+			FormatFields (type, true, accessibility, members, prefix, hash);
+		}
+
+		protected void FormatProperties (TargetClassType type, bool is_static,
+						 TargetMemberAccessibility accessibility,
+						 List<string> members, string prefix, Hashtable hash)
+		{
+			List<TargetPropertyInfo> list = new List<TargetPropertyInfo> ();
+			foreach (TargetPropertyInfo property in type.Properties) {
+				if (property.IsStatic != is_static)
+					continue;
+				if (property.Accessibility != accessibility)
+					continue;
+				list.Add (property);
+			}
+			if (list.Count == 0)
+				return;
+
+			foreach (TargetPropertyInfo property in list)
+				members.Add (FormatProperty (prefix, property, is_static, hash));
+		}
+
+		protected void FormatProperties (TargetClassType type,
+						 TargetMemberAccessibility accessibility,
+						 List<string> members, string prefix, Hashtable hash)
+		{
+			FormatProperties (type, false, accessibility, members, prefix, hash);
+			FormatProperties (type, true, accessibility, members, prefix, hash);
+		}
+
+		protected void FormatEvents (TargetClassType type, bool is_static,
+					     TargetMemberAccessibility accessibility,
+					     List<string> members, string prefix, Hashtable hash)
+		{
+			List<TargetEventInfo> list = new List<TargetEventInfo> ();
+			foreach (TargetEventInfo einfo in type.Events) {
+				if (einfo.IsStatic != is_static)
+					continue;
+				if (einfo.Accessibility != accessibility)
+					continue;
+				list.Add (einfo);
+			}
+			if (list.Count == 0)
+				return;
+
+			foreach (TargetEventInfo einfo in list)
+				members.Add (FormatEvent (prefix, einfo, is_static, hash));
+		}
+
+		protected void FormatEvents (TargetClassType type,
+					     TargetMemberAccessibility accessibility,
+					     List<string> members, string prefix, Hashtable hash)
+		{
+			FormatEvents (type, false, accessibility, members, prefix, hash);
+			FormatEvents (type, true, accessibility, members, prefix, hash);
+		}
+
+		protected void FormatMethods (TargetClassType type, bool is_ctor, bool is_static,
+					      TargetMemberAccessibility accessibility,
+					      List<string> members, string prefix, Hashtable hash)
+		{
+			List<TargetMethodInfo> list = new List<TargetMethodInfo> ();
+			TargetMethodInfo[] methods = is_ctor ? type.Constructors : type.Methods;
+			foreach (TargetMethodInfo method in methods) {
+				if (method.IsStatic != is_static)
+					continue;
+				if (method.Accessibility != accessibility)
+					continue;
+				list.Add (method);
+			}
+			if (list.Count == 0)
+				return;
+
+			foreach (TargetMethodInfo method in list)
+				members.Add (FormatMethod (prefix, method, is_static, is_ctor, hash));
+		}
+
+		protected void FormatMethods (TargetClassType type,
+					      TargetMemberAccessibility accessibility,
+					      List<string> members, string prefix, Hashtable hash)
+		{
+			FormatMethods (type, false, false, accessibility, members, prefix, hash);
+			FormatMethods (type, true, false, accessibility, members, prefix, hash);
+			FormatMethods (type, false, true, accessibility, members, prefix, hash);
+			FormatMethods (type, true, true, accessibility, members, prefix, hash);
+		}
+
+		protected string FormatStruct (string prefix, TargetClassType type, Hashtable hash)
+		{
+			StringBuilder sb = new StringBuilder ();
+
+			List<string> public_members = new List<string> ();
+			List<string> protected_members = new List<string> ();
+			List<string> internal_members = new List<string> ();
+			List<string> private_members = new List<string> ();
+
+			FormatFields (type, TargetMemberAccessibility.Public,
+				      public_members, prefix, hash);
+			FormatFields (type, TargetMemberAccessibility.Protected,
+				      protected_members, prefix, hash);
+			FormatFields (type, TargetMemberAccessibility.Internal,
+				      internal_members, prefix, hash);
+			FormatFields (type, TargetMemberAccessibility.Private,
+				      private_members, prefix, hash);
+
+			FormatProperties (type, TargetMemberAccessibility.Public,
+					  public_members, prefix, hash);
+			FormatProperties (type, TargetMemberAccessibility.Protected,
+					  protected_members, prefix, hash);
+			FormatProperties (type, TargetMemberAccessibility.Internal,
+					  internal_members, prefix, hash);
+			FormatProperties (type, TargetMemberAccessibility.Private,
+					  private_members, prefix, hash);
+
+			FormatEvents (type, TargetMemberAccessibility.Public,
+					  public_members, prefix, hash);
+			FormatEvents (type, TargetMemberAccessibility.Protected,
+					  protected_members, prefix, hash);
+			FormatEvents (type, TargetMemberAccessibility.Internal,
+					  internal_members, prefix, hash);
+			FormatEvents (type, TargetMemberAccessibility.Private,
+					  private_members, prefix, hash);
+
+			FormatMethods (type, TargetMemberAccessibility.Public,
+				       public_members, prefix, hash);
+			FormatMethods (type, TargetMemberAccessibility.Protected,
+				       protected_members, prefix, hash);
+			FormatMethods (type, TargetMemberAccessibility.Internal,
+				       internal_members, prefix, hash);
+			FormatMethods (type, TargetMemberAccessibility.Private,
+				       private_members, prefix, hash);
+
+			if (public_members.Count > 0) {
+				sb.Append (prefix + "public:\n");
+				foreach (string text in public_members)
+					sb.Append (text);
+			}
+
+			if (protected_members.Count > 0) {
+				sb.Append (prefix + "protected:\n");
+				foreach (string text in protected_members)
+					sb.Append (text);
+			}
+
+			if (internal_members.Count > 0) {
+				sb.Append (prefix + "internal:\n");
+				foreach (string text in internal_members)
+					sb.Append (text);
+			}
+
+			if (private_members.Count > 0) {
+				sb.Append (prefix + "private:\n");
+				foreach (string text in private_members)
+					sb.Append (text);
+			}
+
+			return sb.ToString ();
+		}
+
 		protected string PrintObject (Thread target, TargetObject obj)
 		{
 			try {
