@@ -1,8 +1,10 @@
 using System;
 using System.IO;
 using System.Text;
+using System.Reflection;
 using System.Runtime.InteropServices;
 using System.Collections;
+using System.Collections.Generic;
 using System.Threading;
 using C = Mono.CompilerServices.SymbolWriter;
 
@@ -57,6 +59,8 @@ namespace Mono.Debugger.Languages.Mono
 				corlib, memory, FundamentalKind.Boolean);
 			CharType = MonoFundamentalType.Create (
 				corlib, memory, FundamentalKind.Char);
+			ByteType = MonoFundamentalType.Create (
+				corlib, memory, FundamentalKind.Byte);
 			SByteType = MonoFundamentalType.Create (
 				corlib, memory, FundamentalKind.SByte);
 			Int16Type = MonoFundamentalType.Create (
@@ -203,8 +207,7 @@ namespace Mono.Debugger.Languages.Mono
 		Hashtable assembly_hash;
 		Hashtable assembly_by_name;
 		Hashtable class_hash;
-		Hashtable class_info_by_addr;
-		Hashtable class_info_by_type;
+		Dictionary<TargetAddress,MonoClassInfo> class_info_by_addr;
 		MonoSymbolFile corlib;
 		MonoBuiltinTypeInfo builtin_types;
 		MonoFunctionType main_method;
@@ -218,13 +221,11 @@ namespace Mono.Debugger.Languages.Mono
 		MonoDebuggerInfo info;
 		TargetAddress[] trampolines;
 		bool initialized;
-		DebuggerMutex mutex;
 
 		public MonoLanguageBackend (ProcessServant process, MonoDebuggerInfo info)
 		{
 			this.process = process;
 			this.info = info;
-			mutex = new DebuggerMutex ("mono_mutex");
 			data_tables = new Hashtable ();
 		}
 
@@ -325,6 +326,27 @@ namespace Mono.Debugger.Languages.Mono
 				return new MonoPointerType (element_type);
 			}
 
+			Cecil.GenericParameter gen_param = type as Cecil.GenericParameter;
+			if (gen_param != null)
+				return new MonoGenericParameterType (this, gen_param.Name);
+
+			Cecil.GenericInstanceType gen_inst = type as Cecil.GenericInstanceType;
+			if (gen_inst != null) {
+				TargetType[] args = new TargetType [gen_inst.GenericArguments.Count];
+				for (int i = 0; i < args.Length; i++) {
+					args [i] = LookupMonoType (gen_inst.GenericArguments [i]);
+					if (args [i] == null)
+						return null;
+				}
+
+				MonoClassType container = LookupMonoType (gen_inst.ElementType)
+					as MonoClassType;
+				if (container == null)
+					return null;
+
+				return new MonoGenericInstanceType (container, args, TargetAddress.Null);
+			}
+
 			int rank = 0;
 
 			string full_name = type.FullName;
@@ -362,27 +384,59 @@ namespace Mono.Debugger.Languages.Mono
 
 		public TargetType LookupMonoClass (TargetMemoryAccess target, TargetAddress klass_address)
 		{
-			TargetType type = (TargetType) class_hash [klass_address];
-			if (type != null)
-				return type;
+			return ReadClassInfo (target, klass_address).RealType;
+		}
 
-			MonoClassInfo info;
-
-			try {
-				info = MonoClassInfo.ReadClassInfo (this, target, klass_address);
-				if (info == null)
-					return null;
-
-				type = info.SymbolFile.LookupMonoType (info.CecilType);
-			} catch {
-				return null;
+		Cecil.AssemblyDefinition resolve_cecil_asm_name (Cecil.AssemblyNameReference name)
+		{
+			foreach (MonoSymbolFile symfile in symfile_hash.Values) {
+				if (name.FullName == symfile.Assembly.Name.FullName)
+					return symfile.Assembly;
 			}
 
-			if (type == null)
-				return null;
+			return null;
+		}
 
-			class_hash.Add (klass_address, type);
-			return type;
+		Cecil.TypeDefinition resolve_cecil_type_ref (Cecil.TypeReference type)
+		{
+			type = type.GetOriginalType ();
+
+			if (type is Cecil.TypeDefinition)
+				return (Cecil.TypeDefinition) type;
+
+			Cecil.AssemblyNameReference reference = type.Scope as Cecil.AssemblyNameReference;
+			if (reference != null) {
+				Cecil.AssemblyDefinition assembly = resolve_cecil_asm_name (reference);
+				if (assembly == null)
+					return null;
+
+				return assembly.MainModule.Types [type.FullName];
+			}
+
+			Cecil.ModuleDefinition module = type.Scope as Cecil.ModuleDefinition;
+			if (module != null)
+				return module.Types [type.FullName];
+
+			throw new NotImplementedException ();
+		}
+
+		public override bool IsExceptionType (TargetClassType ctype)
+		{
+			MonoClassType mono_type = ctype as MonoClassType;
+			if (mono_type == null)
+				return false;
+
+			Cecil.TypeDefinition exc_type = builtin_types.ExceptionType.Type;
+			Cecil.TypeDefinition type = mono_type.Type;
+
+			while (type != null) {
+				if (exc_type == type)
+					return true;
+
+				type = resolve_cecil_type_ref (type.BaseType);
+			}
+
+			return false;
 		}
 
 		public MonoSymbolFile GetImage (TargetAddress address)
@@ -413,8 +467,7 @@ namespace Mono.Debugger.Languages.Mono
 			assembly_hash = new Hashtable ();
 			assembly_by_name = new Hashtable ();
 			class_hash = new Hashtable ();
-			class_info_by_addr = new Hashtable ();
-			class_info_by_type = new Hashtable ();
+			class_info_by_addr = new Dictionary<TargetAddress,MonoClassInfo> ();
 		}
 
 		void reached_main (TargetMemoryAccess target, TargetAddress method)
@@ -458,6 +511,14 @@ namespace Mono.Debugger.Languages.Mono
 			return target_type;
 		}
 
+		public TargetStructType ReadStructType (TargetMemoryAccess memory, TargetAddress address)
+		{
+			TargetAddress data = MonoRuntime.MonoTypeGetData (memory, address);
+			MonoTypeEnum type = MonoRuntime.MonoTypeGetType (memory, address);
+
+			return (TargetStructType) ReadType (memory, type, data);
+		}
+
 		TargetType ReadType (TargetMemoryAccess memory, MonoTypeEnum type, TargetAddress data)
 		{
 			switch (type) {
@@ -493,6 +554,8 @@ namespace Mono.Debugger.Languages.Mono
 				return BuiltinTypes.IntType;
 			case MonoTypeEnum.MONO_TYPE_U:
 				return BuiltinTypes.UIntType;
+			case MonoTypeEnum.MONO_TYPE_VOID:
+				return BuiltinTypes.VoidType;
 
 			case MonoTypeEnum.MONO_TYPE_PTR: {
 				TargetType target_type = ReadType (memory, data);
@@ -522,12 +585,46 @@ namespace Mono.Debugger.Languages.Mono
 				return new MonoArrayType (etype, rank);
 			}
 
+			case MonoTypeEnum.MONO_TYPE_GENERICINST:
+				return ReadGenericClass (memory, data);
+
+			case MonoTypeEnum.MONO_TYPE_VAR:
+			case MonoTypeEnum.MONO_TYPE_MVAR: {
+				MonoRuntime.GenericParamInfo info = MonoRuntime.GetGenericParameter (
+					memory, data);
+
+				return new MonoGenericParameterType (this, info.Name);
+			}
+
 			default:
 				Report.Error ("UNKNOWN TYPE: {0}", type);
 				return null;
 			}
 		}
 
+		public MonoGenericInstanceType ReadGenericClass (TargetMemoryAccess memory,
+								 TargetAddress address)
+		{
+			MonoRuntime.GenericClassInfo info = MonoRuntime.GetGenericClass (memory, address);
+			if (info == null)
+				return null;
+
+			TargetType[] args = new TargetType [info.TypeArguments.Length];
+
+			for (int i = 0; i < info.TypeArguments.Length; i++) {
+				args [i] = ReadType (memory, info.TypeArguments [i]);
+				if (args [i] == null)
+					return null;
+			}
+
+			MonoClassType container = LookupMonoClass (memory, info.ContainerClass)
+				as MonoClassType;
+			if (container == null)
+				return null;
+
+			return new MonoGenericInstanceType (container, args, info.KlassPtr);
+		}
+
 		internal MonoFunctionType MainMethod {
 			get { return main_method; }
 		}
@@ -536,14 +633,16 @@ namespace Mono.Debugger.Languages.Mono
 		internal void Update (TargetMemoryAccess target)
 		{
 			Report.Debug (DebugFlags.JitSymtab, "Update requested");
-			DateTime start = DateTime.Now;
-			++data_table_count;
-			foreach (MonoDataTable table in data_tables.Values)
-				table.Read (target);
-			foreach (MonoSymbolFile symfile in symfile_by_index.Values)
-				symfile.TypeTable.Read (target);
-			global_data_table.Read (target);
-			data_table_time += DateTime.Now - start;
+			if (initialized) {
+				DateTime start = DateTime.Now;
+				++data_table_count;
+				foreach (MonoDataTable table in data_tables.Values)
+					table.Read (target);
+				foreach (MonoSymbolFile symfile in symfile_by_index.Values)
+					symfile.TypeTable.Read (target);
+				global_data_table.Read (target);
+				data_table_time += DateTime.Now - start;
+			}
 		}
 
 		void read_symbol_table (TargetMemoryAccess memory)
@@ -559,6 +658,10 @@ namespace Mono.Debugger.Languages.Mono
 				update_time += DateTime.Now - start;
 			} catch (ThreadAbortException) {
 				return;
+			} catch (SymbolTableException ex) {
+				Console.WriteLine ("Can't read symbol table: {0} {1}",
+						   memory, ex.Message);
+				return;
 			} catch (Exception e) {
 				Console.WriteLine ("Can't read symbol table: {0} {1} {2}",
 						   memory, e, Environment.StackTrace);
@@ -570,7 +673,8 @@ namespace Mono.Debugger.Languages.Mono
 
 		void read_builtin_types (TargetMemoryAccess memory)
 		{
-			builtin_types = new MonoBuiltinTypeInfo (corlib, memory);
+			if (initialized)
+				builtin_types = new MonoBuiltinTypeInfo (corlib, memory);
 		}
 
 		MonoSymbolFile load_symfile (TargetMemoryAccess memory, TargetAddress address)
@@ -605,12 +709,8 @@ namespace Mono.Debugger.Languages.Mono
 			return symfile;
 		}
 
-		void close_symfile (int index)
+		void close_symfile (MonoSymbolFile symfile)
 		{
-			MonoSymbolFile symfile = (MonoSymbolFile) symfile_by_index [index];
-			if (symfile == null)
-				throw new InternalError ();
-
 			symfile_by_image_addr.Remove (symfile.MonoImage);
 			assembly_hash.Remove (symfile.Assembly);
 			assembly_by_name.Remove (symfile.Assembly.Name.FullName);
@@ -817,15 +917,11 @@ namespace Mono.Debugger.Languages.Mono
 
 		internal MonoClassInfo ReadClassInfo (TargetMemoryAccess memory, TargetAddress klass)
 		{
-			MonoClassInfo info = (MonoClassInfo) class_info_by_addr [klass];
-			if (info == null) {
-				info = MonoClassInfo.ReadClassInfo (this, memory, klass);
-
-				class_info_by_addr.Add (klass, info);
-				if (!info.IsGenericClass)
-					class_info_by_type.Add (info.CecilType, info);
-			}
+			if (class_info_by_addr.ContainsKey (klass))
+				return class_info_by_addr [klass];
 
+			MonoClassInfo info = MonoClassInfo.ReadClassInfo (this, memory, klass);
+			class_info_by_addr.Add (klass, info);
 			return info;
 		}
 
@@ -833,10 +929,9 @@ namespace Mono.Debugger.Languages.Mono
 						       TargetMemoryAccess memory, TargetAddress klass)
 		{
 			MonoClassType type;
-			MonoClassInfo info = MonoClassInfo.ReadClassInfo (
+			MonoClassInfo info = MonoClassInfo.ReadCoreType (
 				file, typedef, memory, klass, out type);
 			class_info_by_addr.Add (klass, info);
-			class_info_by_type.Add (typedef, info);
 
 			return type;
 		}
@@ -905,8 +1000,8 @@ namespace Mono.Debugger.Languages.Mono
 			int index = GetUniqueID ();
 
 			TargetAddress retval = thread.CallMethod (
-				info.InsertSourceBreakpoint, func.MonoClass.File.MonoImage,
-				func.Token, index, func.MonoClass.Name);
+				info.InsertSourceBreakpoint, func.SymbolFile.MonoImage,
+				func.Token, index, func.DeclaringType.BaseName);
 
 			MethodLoadedHandler handler = handle.MethodLoaded;
 
@@ -1066,6 +1161,29 @@ namespace Mono.Debugger.Languages.Mono
 			return new MonoNullObject ((TargetType) type, location);
 		}
 
+		public override TargetObjectObject CreateBoxedObject (Thread thread, TargetObject value)
+		{
+			TargetAddress klass;
+
+			if ((value is MonoClassObject)  && !value.Type.IsByRef)
+				klass = ((MonoClassObject) value).KlassAddress;
+			else if (value is TargetFundamentalObject) {
+				MonoClassType ctype = ((MonoFundamentalType) value.Type).MonoClassType;
+				MonoClassInfo info = (MonoClassInfo) thread.ThreadServant.DoTargetAccess (
+					delegate (TargetMemoryAccess target) {
+						return ctype.ResolveClass (target, true);
+					});
+				klass = info.KlassAddress;
+			} else
+				throw new InvalidOperationException ();
+
+			TargetAddress boxed = thread.CallMethod (MonoDebuggerInfo.GetBoxedObjectMethod, klass, value);
+			if (boxed.IsNull)
+				return null;
+
+			return new MonoObjectObject (builtin_types.ObjectType, new AbsoluteTargetLocation (boxed));
+		}
+
 		public override TargetPointerType CreatePointerType (TargetType type)
 		{
 			return null;
@@ -1190,6 +1308,7 @@ namespace Mono.Debugger.Languages.Mono
 				MonoSymbolFile symfile = load_symfile (inferior, data);
 				Report.Debug (DebugFlags.JitSymtab,
 					      "Module load: {0} {1}", data, symfile);
+				engine.Process.Debugger.OnModuleLoaded (symfile.Module);
 				if ((builtin_types != null) && (symfile != null)) {
 					if (engine.OnModuleLoaded ())
 						return false;
@@ -1197,15 +1316,16 @@ namespace Mono.Debugger.Languages.Mono
 				break;
 			}
 
-			case NotificationType.ReachedMain:
-				if (engine.OnModuleLoaded ())
-					return false;
-				break;
-
 			case NotificationType.UnloadModule:
 				Report.Debug (DebugFlags.JitSymtab,
 					      "Module unload: {0} {1}", data, arg);
-				close_symfile ((int) arg);
+
+				MonoSymbolFile symfile = (MonoSymbolFile) symfile_by_index [(int) arg];
+				if (symfile == null)
+					throw new InternalError ();
+
+				engine.Process.Debugger.OnModuleUnLoaded (symfile.Module);
+				close_symfile (symfile);
 				break;
 
 			case NotificationType.JitBreakpoint:
@@ -1246,11 +1366,11 @@ namespace Mono.Debugger.Languages.Mono
 			}
 
 			if (disposing) {
-				if (symfile_by_index != null) {
-					foreach (MonoSymbolFile symfile in symfile_by_index.Values)
+				if (symfile_hash != null) {
+					foreach (MonoSymbolFile symfile in symfile_hash.Values)
 						symfile.Dispose();
 
-					symfile_by_index = null;
+					symfile_hash = null;
 				}
 			}
 		}
