@@ -9,14 +9,29 @@ namespace Mono.Debugger.Languages.Mono
 	{
 		[NonSerialized]
 		public readonly Cecil.FieldDefinition FieldInfo;
-		public readonly MonoClassType DeclaringType;
 
-		public MonoFieldInfo (MonoClassType type, TargetType field_type, int pos,
+		public MonoFieldInfo (IMonoStructType type, TargetType field_type, int pos,
 				      Cecil.FieldDefinition finfo)
-			: base (field_type, finfo.Name, pos, finfo.IsStatic, pos, 0, finfo.HasConstant)
+			: base (field_type, finfo.Name, pos, finfo.IsStatic,
+				GetAccessibility (finfo), pos, 0, finfo.HasConstant)
 		{
 			FieldInfo = finfo;
-			DeclaringType = type;
+		}
+
+		internal static TargetMemberAccessibility GetAccessibility (Cecil.FieldDefinition field)
+		{
+			switch (field.Attributes & Cecil.FieldAttributes.FieldAccessMask) {
+			case Cecil.FieldAttributes.Public:
+				return TargetMemberAccessibility.Public;
+			case Cecil.FieldAttributes.Family:
+			case Cecil.FieldAttributes.FamANDAssem:
+				return TargetMemberAccessibility.Protected;
+			case Cecil.FieldAttributes.Assembly:
+			case Cecil.FieldAttributes.FamORAssem:
+				return TargetMemberAccessibility.Internal;
+			default:
+				return TargetMemberAccessibility.Private;
+			}
 		}
 
 		public override object ConstValue {
@@ -34,14 +49,14 @@ namespace Mono.Debugger.Languages.Mono
 	{
 		[NonSerialized]
 		public readonly Cecil.FieldDefinition FieldInfo;
-		MonoEnumType type;
 
 		public MonoEnumInfo (MonoEnumType type, TargetType field_type, int index, int pos,
 				     Cecil.FieldDefinition finfo)
-			: base (field_type, finfo.Name, index, finfo.IsStatic, pos, 0, finfo.HasConstant)
+			: base (field_type, finfo.Name, index, finfo.IsStatic,
+				MonoFieldInfo.GetAccessibility (finfo),
+				pos, 0, finfo.HasConstant)
 		{
 			FieldInfo = finfo;
-			this.type = type;
 		}
 
 		public override object ConstValue {
@@ -58,23 +73,37 @@ namespace Mono.Debugger.Languages.Mono
 	internal class MonoMethodInfo : TargetMethodInfo
 	{
 		public readonly MonoFunctionType FunctionType;
-		public readonly MonoClassType Klass;
 
-		private MonoMethodInfo (MonoClassType klass, int index, Cecil.MethodDefinition minfo,
+		private MonoMethodInfo (IMonoStructType klass, int index, Cecil.MethodDefinition minfo,
 					MonoFunctionType type)
 			: base (type, MonoFunctionType.GetMethodName (minfo), index,
-				minfo.IsStatic, type.FullName)
+				minfo.IsStatic, GetAccessibility (minfo), type.FullName)
 		{
-			Klass = klass;
 			FunctionType = type;
 		}
 
-		internal static MonoMethodInfo Create (MonoClassType klass, int index,
+		internal static MonoMethodInfo Create (IMonoStructType klass, int index,
 						       Cecil.MethodDefinition minfo)
 		{
-			MonoFunctionType type = klass.File.LookupFunction (klass, minfo);
+			MonoFunctionType type = klass.LookupFunction (minfo);
 			return new MonoMethodInfo (klass, index, minfo, type);
 		}
+
+		internal static TargetMemberAccessibility GetAccessibility (Cecil.MethodDefinition method)
+		{
+			switch (method.Attributes & Cecil.MethodAttributes.MemberAccessMask) {
+			case Cecil.MethodAttributes.Public:
+				return TargetMemberAccessibility.Public;
+			case Cecil.MethodAttributes.Family:
+			case Cecil.MethodAttributes.FamANDAssem:
+				return TargetMemberAccessibility.Protected;
+			case Cecil.MethodAttributes.Assem:
+			case Cecil.MethodAttributes.FamORAssem:
+				return TargetMemberAccessibility.Internal;
+			default:
+				return TargetMemberAccessibility.Private;
+			}
+		}
 	}
 
 	[Serializable]
@@ -84,9 +113,10 @@ namespace Mono.Debugger.Languages.Mono
 		public readonly MonoFunctionType AddType, RemoveType, RaiseType;
 
 		private MonoEventInfo (MonoClassType klass, int index, Cecil.EventDefinition einfo,
-				       TargetType type, bool is_static,  MonoFunctionType add,
+				       TargetType type, bool is_static,
+				       TargetMemberAccessibility accessibility, MonoFunctionType add,
 				       MonoFunctionType remove, MonoFunctionType raise)
-			: base (type, einfo.Name, index, is_static, add, remove, raise)
+			: base (type, einfo.Name, index, is_static, accessibility, add, remove, raise)
 		{
 			this.Klass = klass;
 			this.AddType = add;
@@ -101,66 +131,75 @@ namespace Mono.Debugger.Languages.Mono
 
 			bool is_static = false;
 			MonoFunctionType add, remove, raise;
+
+			TargetMemberAccessibility accessibility = TargetMemberAccessibility.Private;
 			if (einfo.AddMethod != null) {
-				add = klass.File.LookupFunction (klass, einfo.AddMethod);
+				add = klass.LookupFunction (einfo.AddMethod);
 				is_static = einfo.AddMethod.IsStatic;
+				accessibility = MonoMethodInfo.GetAccessibility (einfo.AddMethod);
 			} else
 				add = null;
 
 			if (einfo.RemoveMethod != null) {
-				remove = klass.File.LookupFunction (klass, einfo.RemoveMethod);
+				remove = klass.LookupFunction (einfo.RemoveMethod);
 				is_static = einfo.RemoveMethod.IsStatic;
+				accessibility = MonoMethodInfo.GetAccessibility (einfo.RemoveMethod);
 			} else
 				remove = null;
 
 			if (einfo.InvokeMethod != null) {
-				raise = klass.File.LookupFunction (klass, einfo.InvokeMethod);
+				raise = klass.LookupFunction (einfo.InvokeMethod);
 				is_static = einfo.InvokeMethod.IsStatic;
+				accessibility = MonoMethodInfo.GetAccessibility (einfo.InvokeMethod);
 			} else
 				raise = null;
 
 			return new MonoEventInfo (
-				klass, index, einfo, type, is_static, add, remove, raise);
+				klass, index, einfo, type, is_static, accessibility, add, remove, raise);
 		}
 	}
 
 	[Serializable]
 	internal class MonoPropertyInfo : TargetPropertyInfo
 	{
-		public readonly MonoClassType Klass;
+		public readonly IMonoStructType Klass;
 		public readonly MonoFunctionType GetterType, SetterType;
 
-		private MonoPropertyInfo (TargetType type, MonoClassType klass, int index,
+		private MonoPropertyInfo (TargetType type, IMonoStructType klass, int index,
 					  bool is_static, Cecil.PropertyDefinition pinfo,
+					  TargetMemberAccessibility accessibility,
 					  MonoFunctionType getter, MonoFunctionType setter)
-			: base (type, pinfo.Name, index, is_static, getter, setter)
+			: base (type, pinfo.Name, index, is_static, accessibility, getter, setter)
 		{
 			this.Klass = klass;
 			this.GetterType = getter;
 			this.SetterType = setter;
 		}
 
-		internal static MonoPropertyInfo Create (MonoClassType klass, int index,
+		internal static MonoPropertyInfo Create (IMonoStructType klass, int index,
 							 Cecil.PropertyDefinition pinfo)
 		{
 			TargetType type = klass.File.MonoLanguage.LookupMonoType (pinfo.PropertyType);
 
 			bool is_static = false;
 			MonoFunctionType getter, setter;
-			if (pinfo.GetMethod != null) {
-				getter = klass.File.LookupFunction (klass, pinfo.GetMethod);
-				is_static = pinfo.GetMethod.IsStatic;
-			} else
-				getter = null;
-
+			TargetMemberAccessibility accessibility = TargetMemberAccessibility.Private;
 			if (pinfo.SetMethod != null) {
-				setter = klass.File.LookupFunction (klass, pinfo.SetMethod);
+				setter = klass.LookupFunction (pinfo.SetMethod);
 				is_static = pinfo.SetMethod.IsStatic;
+				accessibility = MonoMethodInfo.GetAccessibility (pinfo.SetMethod);
 			} else
 				setter = null;
 
+			if (pinfo.GetMethod != null) {
+				getter = klass.LookupFunction (pinfo.GetMethod);
+				is_static = pinfo.GetMethod.IsStatic;
+				accessibility = MonoMethodInfo.GetAccessibility (pinfo.GetMethod);
+			} else
+				getter = null;
+
 			return new MonoPropertyInfo (
-				type, klass, index, is_static, pinfo, getter, setter);
+				type, klass, index, is_static, pinfo, accessibility, getter, setter);
 		}
 	}
 }
