@@ -12,32 +12,32 @@ namespace Mono.Debugger.Languages.Mono
 			this.Type = type;
 		}
 
-		public override TargetType GetCurrentType (Thread target)
+		internal override TargetType GetCurrentType (TargetMemoryAccess target)
 		{
 			return Type.StaticType;
 		}
 
-		public override TargetObject GetDereferencedObject (Thread target)
+		internal override TargetObject GetDereferencedObject (TargetMemoryAccess target)
 		{
-			return Type.StaticType.GetObject (Location);
+			return Type.StaticType.GetObject (target, Location);
 		}
 
-		internal override long GetDynamicSize (Thread target, TargetBlob blob,
+		internal override long GetDynamicSize (TargetMemoryAccess target, TargetBlob blob,
 						       TargetLocation location,
 						       out TargetLocation dynamic_location)
 		{
 			throw new InvalidOperationException ();
 		}
 
-		public override TargetObject GetArrayElement (Thread target, int index)
+		internal override TargetObject GetArrayElement (TargetMemoryAccess target, int index)
 		{
 			throw new InvalidOperationException ();
 		}
 
-		public override string Print (Thread target)
+		internal override string Print (TargetMemoryAccess target)
 		{
 			if (HasAddress)
-				return String.Format ("{0}", Address);
+				return String.Format ("{0}", GetAddress (target));
 			else
 				return String.Format ("{0}", Location);
 		}
