@@ -13,10 +13,22 @@ namespace Mono.Debugger.Languages.Mono
 			this.element_type = element_type;
 		}
 
+		public override bool HasClassType {
+			get { return false; }
+		}
+
+		public override TargetClassType ClassType {
+			get { throw new InvalidOperationException (); }
+		}
+
 		public override bool IsTypesafe {
 			get { return true; }
 		}
 
+		public override bool CanDereference {
+			get { return true; }
+		}
+
 		public override bool HasStaticType {
 			get { return true; }
 		}
@@ -29,7 +41,11 @@ namespace Mono.Debugger.Languages.Mono
 			get { return element_type; }
 		}
 
-		internal override TargetObject GetObject (TargetLocation location)
+		public override bool ContainsGenericParameters {
+			get { return element_type.ContainsGenericParameters; }
+		}
+
+		protected override TargetObject DoGetObject (TargetMemoryAccess target, TargetLocation location)
 		{
 			return new MonoPointerObject (this, location);
 		}
