@@ -1,7 +1,7 @@
 using System;
 using System.Runtime.InteropServices;
 
-using Mono.Debugger.Backends;
+using Mono.Debugger.Backend;
 
 namespace Mono.Debugger.Languages.Native
 {
@@ -20,12 +20,11 @@ namespace Mono.Debugger.Languages.Native
 			this.bfd_container = bfd_container;
 			this.info = info;
 
-			integer_type = new TargetFundamentalType (this, "int", FundamentalKind.Int32, 4);
-			long_type = new TargetFundamentalType (this, "long", FundamentalKind.Int64, 8);
+			integer_type = new NativeFundamentalType (this, "int", FundamentalKind.Int32, 4);
+			long_type = new NativeFundamentalType (this, "long", FundamentalKind.Int64, 8);
 			pointer_type = new NativePointerType (this, "void *", info.TargetAddressSize);
 			void_type = new NativeOpaqueType (this, "void", 0);
 			string_type = new NativeStringType (this, info.TargetAddressSize);
-
 		}
 
 		public override string Name {
@@ -113,13 +112,30 @@ namespace Mono.Debugger.Languages.Native
 		}
 
 		public override TargetObject CreateNullObject (Thread target, TargetType type)
+		{
+			TargetLocation location = new AbsoluteTargetLocation (TargetAddress.Null);
+			NativePointerType pointer = new NativePointerType (this, type);
+			return new NativePointerObject (pointer, location);
+		}
+
+		public override TargetObjectObject CreateBoxedObject (Thread thread, TargetObject value)
 		{
 			throw new NotSupportedException ();
 		}
 
 		public override TargetPointerType CreatePointerType (TargetType type)
 		{
-			return new NativePointerType (this, type.Name + "*", type, type.Size);
+			return new NativePointerType (this, type);
+		}
+
+		public override bool IsExceptionType (TargetClassType type)
+		{
+			return false;
+		}
+
+		internal TargetAddress LookupSymbol (string name)
+		{
+			return bfd_container.LookupSymbol (name);
 		}
 
 		private bool disposed = false;
