@@ -6,15 +6,15 @@ public class Hello : MarshalByRefObject, IHello
 {
 	public void World ()
 	{
-		Console.WriteLine ("Hello World from {0}!", AppDomain.CurrentDomain);
+		Console.WriteLine ("Hello World from {0}!", AppDomain.CurrentDomain); // @MDB LINE: hello
 	}
 }
 
 class X
 {
-	static void Main ()
+	static int Main ()
 	{
-		AppDomain domain = AppDomain.CreateDomain ("Test");
+		AppDomain domain = AppDomain.CreateDomain ("Test"); // @MDB LINE: main
 		Assembly ass = Assembly.GetExecutingAssembly ();
 		Hello hello = (Hello) domain.CreateInstanceAndUnwrap (ass.FullName, "Hello");
 		Console.WriteLine ("TEST: {0}", hello);
@@ -23,9 +23,10 @@ class X
 		Hello bar = new Hello ();
 		bar.World ();
 
-		hello.World ();
+		hello.World ();			// @MDB BREAKPOINT: main2
 		bar.World ();
 
-		AppDomain.Unload (domain);
+		AppDomain.Unload (domain);	// @MDB BREAKPOINT: unload
+		return 0;
 	}
 }
